/*-=-=-=-=-=-=-=-=-=-=-=- RCCL/RCI Notice of Copyright -=-=-=-=-=-=-=-=-=-=-*
 *									    *
 *	Copyright (C) 1987, 1991 by John E. Lloyd.			    *
 *									    *
 *	Information on copyright permissions, as well as a complete	    *
 *	disclaimer of warranty, is given in the file COPYRIGHT.		    *
 *									    *
 *-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=*/

#include <math.h>
#include "../h/cdefs.h"
#include "../h/fatalMsg.h"

#define FSGN(x)		((x) >= 0.0 ? 1.0 : -1.0)

float chasePoint (v, pd, p0, v0, accel, interval)
float *v;
float pd, p0, v0, accel, interval;
{
	float ti, t2, dP, dPi, dPr, a;
	register float t;
	
	ti = FABS(v0)/accel;
	dP = pd - p0;
	dPi = 0.5*v0*ti;
	dPr = dP - dPi;
	a = (dPr >= 0.0) ? accel : -accel;

	if (FSGN(dPr) == FSGN(dPi))
	 { t2 = sqrt(FABS(dPr)/accel + ti*ti);
	   if ((t = t2-ti-interval) >= 0.0)
	    { *v = v0 + interval*a;
	      return (p0 + interval*(v0 + interval*0.5*a));
	    }
	   else if ((t += t2) >= 0.0)
	    { *v = a*t;
	      return (pd - 0.5*a*t*t);
	    }
	   else
	    { *v = 0.0;
	      return (pd);
	    }
	 }
	else
	 { if ((t = ti-interval) >= 0.0)
	    { *v = v0 + interval*a;
	      return (p0 + interval*(v0 + interval*0.5*a));
	    }
	   else
	    { t2 = sqrt(FABS(dPr)/accel);
	      if ((t += t2) >= 0.0)
	       { *v = v0 + interval*a;
		 return (p0 + interval*(v0 + interval*0.5*a));
	       }
	      else if ((t += t2) >= 0.0)
	       { *v = a*t;
		 return (pd - 0.5*a*t*t);
	       }
	      else
	       { *v = 0.0;
		 return (pd);
	       }
	    }
	 }
}

#define EPSILON 1.0e-7
#define ACCEL	1.0
#define SAMPLE	2.0

float testdata[][4] = 
      {
	0.0,	0.0,	0.0,	0.0,
	0.25,	0.0,	0.25,	0.0,
	1.0,	0.0,	1.0,	0.0,
	4.0,	0.0,	2.0,	2.0,
	9.0,	0.0,	2.0,	2.0,
	0.5,	1.0,	0.5,	0.0,
	2.0,	2.0,	2.0,	0.0,
	4.5,	3.0,	4.0,	1.0,
	1.75,	1.0,	1.75,	0.0,
	3.5,	1.0,	3.0,	1.0,
	8.5,	1.0,	4.0,	3.0,
       15.5,	1.0,	4.0,	3.0,
       -0.125,	0.5,   -0.125,	0.0,
	0.25,	1.0,	0.25,	0.0,
       -0.875,	0.5,   -0.75,  -0.5,
       -0.5,	1.0,	0.0,   -1.0,
       -3.5,	1.0,	0.0,   -1.0,
	1.0,	2.0,	2.0,	0.0,

	0.0,	0.0,	0.0,	0.0,
       -0.25,	0.0,   -0.25,	0.0,
       -1.0,	0.0,   -1.0,	0.0,
       -4.0,	0.0,   -2.0,   -2.0,
       -9.0,	0.0,   -2.0,   -2.0,
       -0.5,   -1.0,   -0.5,	0.0,
       -2.0,   -2.0,   -2.0,	0.0,
       -4.5,   -3.0,   -4.0,   -1.0,
       -1.75,  -1.0,   -1.75,	0.0,
       -3.5,   -1.0,   -3.0,   -1.0,
       -8.5,   -1.0,   -4.0,   -3.0,
      -15.5,   -1.0,   -4.0,   -3.0,
        0.125, -0.5,    0.125,	0.0,
       -0.25,  -1.0,   -0.25,	0.0,
        0.875, -0.5,    0.75,   0.5,
        0.5,   -1.0,	0.0,    1.0,
        3.5,   -1.0,	0.0,    1.0,
       -1.0,   -2.0,   -2.0,	0.0,
      };

main(argc, argv)
int argc;
char **argv;
{
	float pd, p0, v0, pc, expected;
	float v;
	int i;
	int testmode = 0;
	int trackmode = 0;

	argc--; argv++;
	if (argc)
	 { if (argc > 1)
	    { usageMsg (-1, "chase <-test> <-track>\n");
	    }
	   else if (strcmp (*argv, "-test") == 0)
	    { testmode = 1;
	    }
	   else if (strcmp (*argv, "-track") == 0)
	    { trackmode = 1;
	    }
	   else
	    { usageMsg (-1, "chase <-test> <-track>\n");
	    }
	 }

	if (testmode)
	 { 
	   for (p0 = -10.0; p0 < 10.0; p0 += 5.0)
	    { for (i=0; i<ARRAY_SIZE(testdata); i++)
	       { pd = p0 + testdata[i][0];
		 v0 = testdata[i][1];
		 pc = chasePoint (&v, pd, p0, v0, ACCEL, SAMPLE);
		 expected = p0 + testdata[i][2];
		 if (FABS(pc - expected) > EPSILON)
		  { printf ("P Error at i=%d pd=%g p0=%g v0=%g: got %g vs. %g\n",
			    i, pd, p0, v0, pc, expected);
		  }
		 expected = testdata[i][3];
		 if (FABS(v - expected) > EPSILON)
		  { printf ("V Error at i=%d pd=%g p0=%g v0=%g: got %g vs. %g\n",
			    i, pd, p0, v0, v, expected);
		  }
	       }
	    }
	 }
	if (trackmode)
	 { p0 = 0.0;
	   v0 = 0.0;
	   while (1)
	    { printf ("pd > ");
	      if (scanf ("%f", &pd) != 1)
	       { printf ("error scanning value\n");
		 break;
	       }
	      pc = chasePoint (&v, pd, p0, v0, 1.0, 1.0);
	      printf ("pc = %g v = %g vobserved = %g\n", pc, v,
		       pc-p0);
	      p0 = pc;
	      v0 = v;
	    }
	   exit (0);
	 }
	while (1)
	 { printf ("pd, p0, v0 > ");
	   if (scanf ("%f%f%f", &pd, &p0, &v0) != 3)
	    { printf ("error scanning values\n");
	      break;
	    }
	   pc = chasePoint (&v, pd, p0, v0, ACCEL, SAMPLE);
	   printf ("pc = %g v = %g\n", pc, v);
	 }
}
