/*-=-=-=-=-=-=-=-=-=-=-=- RCCL/RCI Notice of Copyright -=-=-=-=-=-=-=-=-=-=-*
 *									    *
 *	Copyright (C) 1987, 1991 by John E. Lloyd.			    *
 *									    *
 *	Information on copyright permissions, as well as a complete	    *
 *	disclaimer of warranty, is given in the file COPYRIGHT.		    *
 *									    *
 *-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=*/

#ifdef X86sol_TARGET
#include <unistd.h>
#include <sys/types.h>
#include <sys/mman.h>
#include <fcntl.h>

#include "../h/cdefs.h"
#include "../h/IOdevice.h"
#include "../h/s16dDevice.h"
#include "../h/errorCodes.h"

#include "s16dreg.h"
#include "s16dioctl.h"
#include "rciioctl.h"

#include "../drivers/driErrorMsgs.h"

/* Private driver information relating to an open device */
typedef struct {
	int fd;		/* file descriptor for UNIX driver */
	u_short cmd;	/* The bits written to the csr don't read back, so 
			   we need to keep this shadow copy. In this driver,
			   we need this only for S16D_DISINT. */
} S16dSoft;

#define CYCLES_PER_MS	5000		/* VERY rough estimate, for a SUN 4 */
#define DEFAULT_TIMEOUT	10*CYCLES_PER_MS

/* Send acknowledge to the remote device, using S16D_SIGOUT. To save 
   time, S16D_SIGOUT is TOGGLED, rather than set and cleared, since
   this reduces the number of wait loops required. bitTest is 
   toggled in anticipation of a subsequent call to WAIT_FOR_ACK().
   We should probably do this inside WAIT_FOR_ACK() itself, but if
   it ain't broke ... :) */
#define SEND_ACK() \
	cmdShad ^= S16D_SIGOUT; *csrp = cmdShad; bitTest ^= S16D_SIGIN;

static char *format = "stat=0x%hx ibuf=0x%hx test=0x%hx\n";

/* Wait for acknowledge from the remote device, using S16D_SIGIN.
   S16D_SIGIN is read twice to thwart false readings due to noise
   or cross-talk in the line. As with SEND_ACK(), we wait for S16D_SIGIN
   to simply toggle rather than achieve a prescribed level, which
   reduces the number of wait loops required. */
#define WAIT_FOR_ACK(x) \
	while (((S16D_SIGIN & *csrp) == bitTest || \
		(S16D_SIGIN & *csrp) == bitTest) && --k) \
	   ; \
	if (k == 0) \
	 { errorMsg (0, format, *csrp, s16d->data, bitTest); \
	   errorMsg (ETimeout, driTimeoutMsg, fxnName, (x)); \
	   return (-1); \
	 }

/* Read and write from the data buffer. These are implemented as macros to
   allow us to put in byte swapping, if necessary. */
#define	READ(p) (*p)
#define	WRITE(p,v) (*p = v)

static int s16dInterruptEnable();

s16dRead (dev, buf, size)
IO_DEVICE *dev;
void *buf;
int size;
{
	char *fxnName = "s16dRead()";
	S16dRegs *s16d = (S16dRegs*)dev->location;
	S16dSoft *soft = (S16dSoft*)dev->softstate;

	register volatile u_short *csrp;/* port control/status register */
	register volatile u_short *dbuf;/* S16D data buffer */
	register u_short *dataPtr;	/* where we put the data read */
	register u_short bitTest;	/* expected setting of S16D_SIGIN */
	register u_short cmdShad;	/* used when writing to s16d->csr */
	register int count;		/* word count register	*/
	register int i;

	short myChecksum;		/* checksum computed at this end */
	short theirChecksum;		/* checksum sent from the other end */
	unsigned long k;		/* timeout counter */
	int byteCount;			/* number of bytes read in */

	dbuf = &(s16d->data);
	csrp = &(s16d->csr);

	/* get initial S16D_SIGIN setting, and initialize cmdShad */
	bitTest = (*csrp & S16D_SIGIN);
	cmdShad = soft->cmd;
	if (*csrp & S16D_SIGOUT_R)
	 { cmdShad |= S16D_SIGOUT;
	 }

	/* -- announce we're ready to read, and then wait for an ACK -- */

	WRITE(dbuf, S16D_READ_READY);
	k = dev->blocktime * CYCLES_PER_MS * 5;
	WAIT_FOR_ACK(1);

	/* -- clear output buffer, and then read no. of words to xfer -- */

	WRITE (dbuf, 0);
	count = (int)READ(dbuf);
	dev->byteCount = count*sizeof(short);

	/* -- check for bogus count: negative, or too big -- */

	if (count < 0 || size < dev->byteCount)
	 { errorMsg (ECount, driCountMsg, fxnName, count);
	   *csrp = soft->cmd; /* clear SIGOUT */
	   return (-1);
	 }

	/* -- read in all the words -- */

	k = (count+2)*dev->timeout;

	dataPtr = (u_short *)buf;
	myChecksum = 0;
	i = count;
	while (i--)
         { SEND_ACK();
	   WAIT_FOR_ACK(2);
	  	
	   *dataPtr = READ(dbuf);
	   myChecksum += *dataPtr++;
	 }

	/* -- get their checksum, and acknowledge it -- */

	SEND_ACK();
	WAIT_FOR_ACK(3);
	theirChecksum = READ(dbuf);
	SEND_ACK();

	/* -- finish handshake -- */

	WAIT_FOR_ACK(4);

	/* -- input buffer is now clear. Verify checksum -- */

	if (myChecksum != theirChecksum)
	 { errorMsg (ETransmit, driChecksumMsg, 
				fxnName, theirChecksum, myChecksum);
	   rciPrintf ("s16dRead: bad checksum. Count = %d\n", count);
	   for (i=0; i<count; i++)
	    { rciPrintf ("%2d: 0x%4x\n", i, ((u_short*)buf)[i] );
	    }
	   return (-1);
	 }
	return (dev->byteCount);
}

s16dWrite (dev, buf, size)
IO_DEVICE *dev;
void *buf;
int size;
{
	char *fxnName = "s16dWrite()";
	S16dRegs *s16d = (S16dRegs*)dev->location;
	S16dSoft *soft = (S16dSoft*)dev->softstate;

	register volatile u_short *csrp;/* port control/status register */
	register volatile u_short *dbuf;/* S16D data buffer */
	register u_short *dataPtr;      /* where we get data to write */
	register u_short bitTest;	/* expected setting of S16D_SIGIN */
	register u_short cmdShad;    	/* used when writing to s16d->csr */
	register int count;		/* word count register */

	short checksum;			/* computes checksum */
	unsigned long k;		/* timeout counter */
	int byteCount;			/* number of bytes written */

	dbuf = &(s16d->data);
	csrp = &(s16d->csr);

	/* initialize cmdShad */
	cmdShad = soft->cmd;
	if (*csrp & S16D_SIGOUT_R)
	 { cmdShad |= S16D_SIGOUT;
	 }

	/* -- check for bogus count -- */

        count = size/sizeof(short);
	if (size % sizeof(short))
	 { count++;
	 }
        if (count < 0)
         { errorMsg (ECount, driCountMsg, fxnName, count);
	   *dbuf = 0;
	   *csrp = soft->cmd;	/* clear SIGOUT */
	   return (-1);
	 }

	/* -- announce we're ready to write, then wait for S16D_READ_READY -*/
	WRITE(dbuf, S16D_WRITE_READY);
	k = dev->blocktime * CYCLES_PER_MS;
	while (READ(dbuf) != S16D_READ_READY && --k)
	   ;
	if (k == 0)
	 { errorMsg (ETimeout, driTimeoutMsg, fxnName, 0);
	   *dbuf = 0;
	   *csrp = soft->cmd;	/* clear SIGOUT */
	   return (-1);
	 }
	
	/* -- write out the number of words to be written -- */

        WRITE(dbuf,count);

	/* get initial setting of S16D_SIGIN, and then toggle it to
	   counteract subsequent toggling by SEND_ACK() */
	bitTest = (*csrp & S16D_SIGIN) ^ S16D_SIGIN;

	k = (count+2)*dev->timeout;

	/* -- handshake after writing out the count -- */

	SEND_ACK();
	WAIT_FOR_ACK(1);

	dataPtr = (u_short *)buf;
	
	/* -- send words down the chute -- */

	checksum = 0;
	while (count--)
	 { checksum += *dataPtr;
	   WRITE (dbuf, *dataPtr++);	/* load the s16d */

	   SEND_ACK();
	   WAIT_FOR_ACK(2);
         }

	/* -- send our checksum -- */

	WRITE (dbuf, checksum);
	SEND_ACK();
	WAIT_FOR_ACK(3);

	*dbuf = 0;
	SEND_ACK();

	/* make sure signal bit is clear before we leave routine */

	if (*csrp & S16D_SIGOUT_R)
	 { WAIT_FOR_ACK(4);
	   SEND_ACK();
	 }
	k = 1000;
	while (--k)
	   ;

	return (size);
}

s16dProbe (dev)
IO_DEVICE *dev;
{
	/* Check to see if the device is there. We do this by trying
	   to open whatever special file is indicated by addrStr. This
	   UNIX device is then used to get a memory map of the device
	   registers */

	char *fxnName = "s16dProbe()";
	S16dRegs *s16d;
	S16dSoft *soft;
	u_short ushort;
	int fd;

	/* Variables used to map the device registers into memory */

	unsigned long regAddr = 0;
	int pageSize;
	off_t pageStart;
	caddr_t vpageStart;
	u_long pageOffset;


	if ((fd = open (dev->addrStr, O_RDWR)) < 0)
	 { sysErrorMsg (EConnect, fxnName, "open");
	   return (-1);
	 }

	/* map in the device registers */

#ifdef __STDC__
    	pageSize = sysconf(_SC_PAGESIZE);
#else
    	pageSize = getpagesize();
#endif
	pageStart = (regAddr / pageSize) * pageSize;
	pageOffset = regAddr % pageSize;
	vpageStart = mmap((caddr_t)0, sizeof(S16dRegs), PROT_READ|PROT_WRITE,
			   MAP_SHARED, fd, pageStart);
	if (vpageStart == (caddr_t)(-1))
	 { sysErrorMsg (EConnect, fxnName, "mmap");
	   close (fd);
	   return (-1);
	 }

	/* map has been obtained. Save it. */

	dev->location = vpageStart + pageOffset;
	s16d = (S16dRegs*)dev->location;

	/* Issue a reset command to the board */

	s16d->csr = S16D_RESET;

	/* enable byte swapping on input/output. We do this using
	   ioctls because the config register state is stored by
	   the UNIX driver. */

	if (ioctl (fd, S16D_GET_CONFIG, &ushort) < 0)
	 { sysErrorMsg (EConnect, fxnName, "ioctl S16D_GET_CONFIG");
	   close (fd);
	   return (-1);
	 }
	ushort |= S16D_SWAP;
	if (ioctl (fd, S16D_SET_CONFIG, &ushort) < 0)
	 { sysErrorMsg (EConnect, fxnName, "ioctl S16D_SET_CONFIG");
	   close (fd);
	   return (-1);
	 }

	/* get the current contents of the cmd register, so that
	   we can set soft->cmd */

	if (ioctl (fd, S16D_GET_COMMAND, &ushort) < 0)
	 { sysErrorMsg (EConnect, fxnName, "ioctl S16D_GET_CMD");
	   close (fd);
	   return (-1);
	 }

	soft = (S16dSoft*)malloc(sizeof(S16dSoft));
	soft->fd = fd;

	/* in this driver, we only shadow DISINT (disable interrupts) */
	soft->cmd = (ushort & S16D_DISINT);
	dev->softstate = soft;

	dev->timeout = DEFAULT_TIMEOUT;
	dev->blocktime = S16D_DEFAULT_BLOCKTIME;
	dev->size = sizeof(S16dRegs);

	return (sizeof(S16dRegs));
}

s16dAttach (dev)
IO_DEVICE *dev;
{
	/* Initiate connection with robot on the other end. */

	char *fxnName = "s16dAttach()";
	S16dRegs *s16d = (S16dRegs *)dev->location;
	S16dSoft *soft = (S16dSoft *)dev->softstate;
	unsigned long k;

	/* -- check that the other end is there and listening -- */

	if (READ(&s16d->data) != S16D_LISTEN)
	 { errorMsg (EConnect, driListeningMsg, fxnName);
	   return (-1);
	 }

	/* -- announce that we're also listening now -- */

	WRITE(&s16d->data, S16D_LISTEN);

	/* -- clear and set S16D_SIGOUT, which *at this point* should -- */
	/* -- result in an interrupt being sent to the other side -- */
	
	s16d->csr = soft->cmd;
	s16d->csr = soft->cmd | S16D_SIGOUT;

	/* -- wait for other side to clear its output register -- */
	
	k = dev->timeout*100;
	while (READ(&s16d->data) && --k)
	   ;
	if (k == 0)
	 { errorMsg (ETimeout, driTimeoutMsg, fxnName, 1);
	   return (-1);
	 }

	s16d->csr = soft->cmd; /* clear SIGOUT */

	/* -- now clear our output register, and wait for an ACK -- */

	WRITE(&s16d->data,0);
	k = dev->timeout*100;
	while ((s16d->csr & S16D_SIGIN) && --k)
	   ;
	if (k == 0)
	 { errorMsg (ETimeout, driTimeoutMsg, fxnName, 1);
	   return (-1);
	 }

	return (1);	
}

s16dDetach (dev)
IO_DEVICE *dev;
{
	if (dev->location)
	 { S16dRegs *s16d = (S16dRegs*)dev->location;
	   S16dSoft *soft = (S16dSoft*)dev->softstate;

	   WRITE (&s16d->data,0);
	   s16dInterruptEnable (soft, /*enable=*/0);
	   close (soft->fd);
	   free (soft);
	   dev->softstate = NULL;
	   return (1);
	 }
	else
	 { return (0);
	 }
}

s16dSignal (dev, n)
IO_DEVICE *dev;
int n;
{
	/* Send a signal to the other side (normally used to announce 
	   the start of a new control cycle, or an abnormal shutdown) */

	char *fxnName = "s16dSignal()";
	S16dRegs *s16d = (S16dRegs*)dev->location;
	S16dSoft *soft = (S16dSoft*)dev->softstate;
	u_short cmdShad;
       	int k;
	int retval = 1;

	/* check that signal value is 1 or 2 */

	if (n < 1 || n > 2)
	 { errorMsg (ECount, driCountMsg, fxnName, n);
	   return (-1);
	 }

	/* initialize cmdShad */
	cmdShad = soft->cmd;
	if (s16d->csr & S16D_SIGOUT_R)
	 { cmdShad |= S16D_SIGOUT;
	 }

	/* -- sanity check: make sure other side isn't already -- */
	/* -- outputting S16D_SIGNAL_ACK -- */

	if (READ(&s16d->data) == S16D_SIGNAL_ACK)
	 { errorMsg (ESync, driSyncMsg, fxnName);
	   return (-1);
	 }

	/* -- write out signal number, then interrupt other side -- */

	WRITE (&s16d->data, n);
	s16d->csr = cmdShad | S16D_INTR_SET; /* send interrupt */

	/* -- wait for S16D_SIGNAL_ACK -- */

	k = dev->timeout * 50;
	while (READ(&s16d->data) != S16D_SIGNAL_ACK && --k)
	   ;
	if (k == 0)
	 { errorMsg (ETimeout, driTimeoutMsg, fxnName, 1);
	   retval = -1;
	 }
	s16d->csr = cmdShad; /* clear interrupt */

	/* -- now clear our output register -- */
	WRITE (&s16d->data, 0);
	if (retval == -1)
	 { return (-1);
	 }

	/* -- wait for other side to clear S16D_SIGNAL_ACK -- */
	k = dev->timeout * 100;
	while (READ(&s16d->data) == S16D_SIGNAL_ACK && --k)
	   ;
	if (k == 0)
	 { errorMsg (ETimeout, driTimeoutMsg, fxnName, 2);
	   return (-1);
	 }

	return (1);
}

s16dIoctl (dev, com, data)
IO_DEVICE *dev;
int com;
void *data;
{
	char *fxnName = "s16dIoctl()";
	S16dRegs *s16d = (S16dRegs*)dev->location;
	S16dSoft *soft = (S16dSoft*)dev->softstate;

	switch (com)
	 { case IOC_SIGINTER_ENABLE:
	    { s16dInterruptEnable (soft, /*enable=*/1);
	      break;
	    }
	   case IOC_SIGINTER_DISABLE:
	    { s16dInterruptEnable (soft, /*enable=*/0);
	      break;
	    }
	   case IOC_TIMEOUT:
	    { dev->timeout = *(int *)data;
	      break;
	    }
	   case IOC_GET_BLOCKTIME:
	    { *(int *)data = dev->blocktime;
	      break;
	    }
	   case IOC_SET_BLOCKTIME:
	    { dev->blocktime = *(int *)data;
	      break;
	    }
	   default:
	    { errorMsg (EBadIoctl, driIoctlMsg, fxnName, com);
	      return (-1);
	    }
	 }
	return (1);
}

s16dTestRead (dev)
IO_DEVICE *dev;
{
	S16dRegs *s16d = (S16dRegs*)dev->location;

	if (READ(&s16d->data) == S16D_WRITE_READY)
	 { return (1);
	 }
 	else
	 { return (0);
	 }
}

s16dTestWrite (dev)
IO_DEVICE *dev;
{
	S16dRegs *s16d = (S16dRegs*)dev->location;

	if (READ(&s16d->data) == S16D_READ_READY)
	 { return (1);
	 }
 	else
	 { return (0);
	 }
}

static s16dInterruptEnable (soft, enable)
S16dSoft *soft;
int enable;
{
	if (( enable &&  (soft->cmd & S16D_DISINT)) ||
	    (!enable && !(soft->cmd & S16D_DISINT)))
	 { if (ioctl (soft->fd, RCIIOC_INTR_ENABLE, enable) < 0)
	    { perror ("FATAL: s16dInterruptEnable, ioctl");
	      exit (1);
	    }
	 }
	if (enable)
	 { soft->cmd &= ~S16D_DISINT;
	 }
	else
	 { soft->cmd |= S16D_DISINT;
	 }
}
#endif

