/*-=-=-=-=-=-=-=-=-=-=-=- RCCL/RCI Notice of Copyright -=-=-=-=-=-=-=-=-=-=-*
 *									    *
 *	Copyright (C) 1987, 1991 by John E. Lloyd.			    *
 *									    *
 *	Information on copyright permissions, as well as a complete	    *
 *	disclaimer of warranty, is given in the file COPYRIGHT.		    *
 *									    *
 *-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=*/

#if defined(Sun4sol_TARGET) && defined(UBC_CS)

#include "../h/cdefs.h"
#include "../h/IOdevice.h"
#include "../h/errorCodes.h"
#include "../h/errorStack.h"
#include "../h/tlkDevice.h"
#include "../h/rciTask.h"

#include "../drivers/driErrorMsgs.h"

#define DEFAULT_TIMEOUT		1 /* seconds */

/*
 * Link interface functions, taken from 
 * 	/lci/project/transputer/link/bpplink.c
 * 
 * Stewart Kingdon,
 * Copyright (c) 1995,
 * Laboratory for Computational Intelligence,
 * University of British Columbia.
 */

#include <stdio.h>
#include <stdlib.h>
#include <sys/types.h>
#include <fcntl.h>
#include <sys/bpp_io.h> 
#include <sys/ioctl.h>
#include <sys/ioccom.h>
#include <errno.h>

#define LINKNAME	"/dev/bpp0"

static testrc = 0;

static int BPPOpenLink( char *name )
{
    int  link;
    struct bpp_transfer_parms	parms;

    if( !name )
	name = LINKNAME;
    else if( strncmp( name, LINKNAME, sizeof(LINKNAME)-2 ) != 0 )
	return( -1001 );
    link = open( name, O_RDWR );

    if( link < 0 )
	 { printf ("Errno=%d\n", errno);
	   return( -1002 );
	 }
    if( ioctl( link, BPPIOC_GETPARMS, &parms ) < 0 )
	goto ioctl_failed;
    parms.write_timeout = 1;
    parms.read_timeout = 1;
    parms.read_strobe_width = 500;
    parms.read_setup_time = 500;
    if( ioctl( link, BPPIOC_SETPARMS, &parms ) < 0 )
	goto ioctl_failed;
    if( 1 ) {
	char *tmp = (char*)getenv( "BPPTESTRC" );
	if( tmp )
	    testrc = atoi(tmp);
    }
    return( link );

  ioctl_failed:
    close( link );
    return( -1003 );
}

static int BPPCloseLink( int link )
{
    if( link < 0 || close( link ) < 0 )
	return( -1 );
    return( 1 );
}

static int BPPWriteLink( int link, void *_buf, unsigned len, int timeout )
{
    int n = 0,
	t = 0,
	nb;

    while( n < len && (timeout == 0 || t++ < timeout) ) {
	nb = write( link, n + (char*)_buf, len - n );
	if( nb < 0 )
	    return( n ? n : nb );
	n += nb;
    }

    return( n );
}

static int BPPReadLink( int link, void *_buf, unsigned len, int timeout )
{
    int n = 0,
	t = 0,
	nb;

    while( n < len && (timeout == 0 || t++ < timeout) ) {
	nb = read( link, n + (char*)_buf, len - n );
	if( nb < 0 )
	    return( n ? n : nb );
	n += nb;
    }

    return( n );
}

/* -=--=--=--=--=--=--=--=--=--=--=--=--=--=--=--=--=--=--=--=--=--=--=--=- */

doReadLink (ld, buf, size, timeout)
int ld;
unsigned char *buf;
int size;
int timeout;
{
	int status;
	static unsigned char lastBuf1[256];
	static unsigned char lastBuf2[256];
	static unsigned char lastBuf3[256];

	rciSaveContext();
	status = BPPReadLink (ld, buf, size, timeout);
	rciRestoreContext();
	if (status != size)
	 { if (status == -1)
	    { sysErrorMsg (ETransmit, "ReadLink", "");
	    }
	   else
	    { errorMsg (ETransmit, "ReadLink() -- returns %d vs. %d\n",
				    status, size);
	       { int i;
		 printf ("\n   NEW OLD1 OLD2 OLD3\n");
		 for (i=0; i<size; i++)
		  { printf ("%2d  %02x   %02x   %02x   %02x\n", 
			       i, buf[i], lastBuf1[i], lastBuf2[i], lastBuf3[i]);
		  }
	       }
	    }
	   size = -1;
	 }
	memcpy (lastBuf3, lastBuf2, size);
	memcpy (lastBuf2, lastBuf1, size);
	memcpy (lastBuf1, buf, size);
	return (size);
}

doWriteLink (ld, buf, size, timeout)
int ld;
unsigned char *buf;
int size;
int timeout;
{
	int status;

	rciSaveContext();
	status = BPPWriteLink (ld, buf, size, timeout);
	rciRestoreContext();
	if (status != size)
	 { if (status == -1)
	    { sysErrorMsg (ETransmit, "WriteLink", "");
	    }
	   else
	    { errorMsg (ETransmit, "WriteLink() -- returns %d vs. %d\n",
				    status, size);
	    }
	   size = -1;
	 }
	return (size);
}

tlkRead (dev, buf, size)
IO_DEVICE *dev;
void *buf;
int size;
{
	char *fxnName = "tlkRead()";
	struct tlkreg *tlkdev = (struct tlkreg *)dev->location;
	char headbuf[1];

	if (doReadLink (tlkdev->fd, headbuf, 1, dev->timeout) < 0)
	 { errorMsg (getErrorCode(0), "%s -- location A\n", fxnName);
	   return -1;
	 }
	if (headbuf[0])
	 { if (doReadLink (tlkdev->fd, buf, headbuf[0], dev->timeout) < 0)
	    { errorMsg (getErrorCode(0), "%s -- location B\n", fxnName);
	      return -1;
	    }
	 }
	dev->byteCount = headbuf[0];
	return (dev->byteCount);
}

tlkWrite (dev, buf, size)
IO_DEVICE *dev;
void *buf;
int size;
{
	/* send a buffer of 'size' bytes to the target device. Block
	   until the target device is ready to read (or we timeout). */
	   
	char *fxnName = "tlkWrite()";
	struct tlkreg *tlkdev = (struct tlkreg *)dev->location;

	if (doWriteLink (tlkdev->fd, buf, size, dev->timeout) < 0)
	 { errorMsg (getErrorCode(0), "%s -- location A\n", fxnName);
	   return -1;
	 }
	return (size);
}

#include <fcntl.h>	/* Solaris need this for O_RDWR */

tlkAttach(dev)
IO_DEVICE *dev;
{
	struct tlkreg *tlkdev;
	int fd;

	if (dev->location)
	 { tlkdev = (struct tlkreg *)dev->location;
	   if (tlkdev->fd != -1)
	    { BPPCloseLink(tlkdev->fd);
	    }
	   free (tlkdev);
	 }
	if ((fd = BPPOpenLink(dev->addrStr)) < 0)
	 { sysErrorMsg (EConnect, "tlkAttach", "open");
	   return (-1);
	 }
	tlkdev = ALLOC(struct tlkreg);
	tlkdev->fd = fd;
	dev->location = (unsigned char *)tlkdev;
	dev->timeout = DEFAULT_TIMEOUT;
	return (1);
}

tlkDetach (dev)
IO_DEVICE *dev;
{
	struct tlkreg *tlkdev;

	if (dev->location)
	 { tlkdev = (struct tlkreg *)dev->location;
	   if (tlkdev->fd != -1)
	    { BPPCloseLink (tlkdev->fd);
	    }
	   free (tlkdev);
	   dev->location = 0;
	   return (1);
	 }
	else
	 { return (0);
	 }
}

tlkIoctl (dev, com, data)
IO_DEVICE *dev;
int com;
void *data;
{
	char *fxnName = "tlkIoctl()";

	/* Controls various I/O parameters and also allows interrupts to
	   be enabled or disabled */

	struct tlkreg *tlkdev = (struct tlkreg *)dev->location;

	switch (com)
	 { case IOC_SIGINTER_ENABLE:
	    { break;
	    }
	   case IOC_SIGINTER_DISABLE:
	    { break;
	    }
	   case IOC_TIMEOUT:
	    { dev->timeout = *(int *)data;
	      break;
	    }
	   case IOC_GET_BLOCKTIME:
	    { *(int *)data = dev->timeout;
	      break;
	    }
	   case IOC_SET_BLOCKTIME:
	    { dev->timeout = *(int *)data;
	      break;
	    }
	   default:
	    { errorMsg (EBadIoctl, driIoctlMsg, fxnName, com);
	      return (-1);
	    }
	 }
	return (1);
}

tlkTestRead (dev)
IO_DEVICE *dev;
{
	rciAbort (ENotSupported, "tlkTestRead() -- not implemented\n");
	return (0);
}

tlkTestWrite (dev)
IO_DEVICE *dev;
{
	rciAbort (ENotSupported, "tlkTestWrite() -- not implemented\n");
	return (0);
}

#endif
