/*-=-=-=-=-=-=-=-=-=-=-=- RCCL/RCI Notice of Copyright -=-=-=-=-=-=-=-=-=-=-*
 *									    *
 *	Copyright (C) 1987, 1991 by John E. Lloyd.			    *
 *									    *
 *	Information on copyright permissions, as well as a complete	    *
 *	disclaimer of warranty, is given in the file COPYRIGHT.		    *
 *									    *
 *-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=*/

#include <stdio.h>
#include <math.h>

#include "../h/cdefs.h"
#include "../h/robotClass.h"
#include "../h/elbow_kynvar.h"
#include "../h/errorCodes.h"
#include "../h/errorStack.h"
#include "../h/robotMath.h"

#define EPSILONF 	1.0e-7

elbow_kyn (kyn, fp)
KYN *kyn;
FILE *fp;
{
	ELBOW_KYN *ekyn = (ELBOW_KYN*)kyn->ext;
	ConfCodeDesc *desc;

	float a2, a3, a4;
	float cp20, cp30, cp40, cp50;

	char *fxnName = "elbow_kyn()";

	readInit(fp,"=");
	if (!readExactString("A2") || !readExactString("=") ||
	    !readFloat(&a2) ||
	    !readExactString("A3") || !readExactString("=") ||
	    !readFloat(&a3) ||
	    !readExactString("A4") || !readExactString("=") || 
	    !readFloat(&a4))
	 { errorMsg (EBadFile, 
"%s -- can't read `A' matrix parameters\n", fxnName);
	   return (-1);
	 }
	if (!readExactString("CP20") || !readExactString("=") ||
	    !readFloat(&cp20) ||
	    !readExactString("CP30") || !readExactString("=") ||
	    !readFloat(&cp30) ||
	    !readExactString("CP40") || !readExactString("=") || 
	    !readFloat(&cp40) ||
	    !readExactString("CP50") || !readExactString("=") || 
	    !readFloat(&cp50))
	 { errorMsg (EBadFile, 
"%s -- can't read gravity coefficients\n", fxnName);
	   return (-1);
	 }
	
	kyn->gen->dh[1].a = a2;
	kyn->gen->dh[2].a = a3;
	kyn->gen->dh[3].a = a4;

	kyn->gen->dh[3].alpha =  PI/2;
	kyn->gen->dh[4].alpha = -PI/2;

	ekyn->cp20 = cp20;
	ekyn->cp30 = cp30;
	ekyn->cp40 = cp40;
	ekyn->cp50 = cp50;	

	/* now read the friction values */

	{ static char *labels[3] = { "STATF", "COULF", "VISCF" };
	  float *vector[3];
	  char fmt[64];
	  int k, j;

	  vector[0] = kyn->gen->statFriction;
	  vector[1] = kyn->gen->coulFriction;
	  vector[2] = kyn->gen->viscFriction;
	  for (k=0; k<3; k++)
	   { for (j=0; j<ELBOWJ; j++)
	      { sprintf (fmt, "%s%d", labels[k], j+1);
		if (!readExactString(fmt) || !readExactString("=") ||
		    !readFloat (&((vector[k])[j])) )
		 { errorMsg (EBadFile,
"%s -- can't read %s for joint %d\n", fxnName, labels[k], j+1);
		   return (-1);
		 }
	      }
	   }
	}

	if (gen_readNomVelAcc (fxnName, kyn->gen) == -1)
	 { return (-1);
	 }

	/* ---- Configuration codes ---- */

	kyn->gen->numConf = 3;
	desc = kyn->gen->confCodes;
	desc[0].code  = ELBOW_RIGHT_CONF;
	desc[0].set   = 'r';
	desc[0].clear = 'l';
	desc[1].code  = ELBOW_DOWN_CONF;
	desc[1].set   = 'd';
	desc[1].clear = 'u';
	desc[2].code  = ELBOW_FLIP_CONF;
	desc[2].set   = 'f';
	desc[2].clear = 'n';
	
	/* ---- Instantiate function fields ---- */

	kyn->gen->fwdKinematics = elbowFwdKinematics;
	kyn->gen->invKinematics = elbowInvKinematics;
	kyn->gen->rangeAngles = elbowRangeAngles;
	kyn->gen->solveConf = elbowSolveConf;
	kyn->gen->fwdJacob = elbowFwdJacob;
	kyn->gen->invJacob = elbowInvJacob;
	kyn->gen->fwdJacobT = elbowFwdJacobT;
	kyn->gen->invJacobT = elbowInvJacobT;
	kyn->gen->gravload = elbowGravload;
	kyn->gen->fwdJacobVar = elbowFwdJacobVar;
	kyn->gen->invJacobVar = elbowInvJacobVar;
	kyn->gen->fwdJacobTVar = elbowFwdJacobTVar;
	kyn->gen->invJacobTVar = elbowInvJacobTVar;
	kyn->gen->gravloadVar = elbowGravloadVar;
	kyn->gen->updateVar = elbowUpdateVar;
	kyn->gen->updateVarXsincos = elbowUpdateVarXsincos;

	kyn->gen->rbtclass = ELBOW;

	return (1);
}
