#ifdef X86sol_TARGET
#include <stdio.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <sys/asynch.h>
#include <sys/time.h>
#include <errno.h>
#include <cdefs.h>
#include <unistd.h>

#include "../h/termline.h"
#include "panTilt.h"

static TermProps oldLineState;

#define STX	0x02
#define ETX	0x03
#define ADDR_0	0x00
#define ADDR_1	0x01

#define NO_REPLY	1
#define REPLY		0
#define SHORT_COMSET	0
#define LONG_COMSET	1

#define GET_POS		0x41
#define SET_POS_VEL	0x1a
#define SET_POS		0x19
#define SET_COMSET	0x82
#define SET_COORDINATES	0x3f
#define GET_COORDINATES	0x80
#define SET_ADDRESS	0x0b
#define SET_BAUD	0x0a
#define CHECK_BAUD	0x0d
#define CALIBRATE	0x3c
#define VERIFY_POS	0x3d
#define GET_REFER_POS	0x83
#define STOP		0x14
#define GET_BASE_CFG	0x8c
#define SET_BASE_CFG	0x8d
#define GET_HEAD_CFG	0x8f
#define SET_HEAD_CFG	0x90
#define GET_SYS_STATE   0x01

#define TABLETOP	0
#define SIDEWAYS	4
#define UPSIDE_DOWN	1

#define BAUD_2400	3
#define BAUD_4800	4
#define BAUD_9600	5
#define BAUD_19200	6
#define BAUD_38400	7

#define MAX_RATE        1000

union spack {
	short s;
	char c[2];
};

static int computeCRC(unsigned char *data, int size)
{
	int i;
	unsigned int CRC_val = 0x00;

	for(i=0; i<size; i++)
	 { CRC_val = (CRC_val>>4)^(((CRC_val^data[i]) & 0x0f) * 0x1081);
	   CRC_val = (CRC_val>>4)^(((CRC_val^(data[i]>>4)) & 0x0f)*0x1081);
	 }
	return (unsigned char)(CRC_val & 0x000000ff);
}

int panTiltOpen (char *name)
{
	TermProps newLineState;
	int fd, itry;
	int infoBuf[64];
	char state;

	if ((fd = open(name, O_RDWR)) == -1)
	 { perror ("panTiltOpen");
	   return -1;
	 }
	termlineGet (fd, &oldLineState);
	newLineState = oldLineState;
	termlineRawMode (&newLineState);
	termlineEchoOff (&newLineState);

  newLineState.data.c_lflag = 0;
  newLineState.data.c_oflag = 0;

#if 0
  newLineState.data.c_cc[VINTR] = 0;
  newLineState.data.c_cc[VQUIT] = 0;
  newLineState.data.c_cc[VERASE] = 0;
  newLineState.data.c_cc[VKILL] = 0;
  newLineState.data.c_cc[VEOF] = 0;
  newLineState.data.c_cc[VEOL] = 0;
  newLineState.data.c_cc[VEOL2] = 0;
  newLineState.data.c_cc[VSWTCH] = 0;
  newLineState.data.c_cc[VSUSP] = 0;
  newLineState.data.c_cc[VREPRINT] = 0;
  newLineState.data.c_cc[VDISCARD] = 0;
  newLineState.data.c_cc[VWERASE] = 0;
  newLineState.data.c_cc[VLNEXT] = 0;
  /* missing from rod's code... */
  newLineState.data.c_cc[VSTART] = 0;
  newLineState.data.c_cc[VSTOP] = 0;
#else
  memset( newLineState.data.c_cc, 0, sizeof(newLineState.data.c_cc) );
#endif

	termlineSet (fd, &newLineState);
	
	if (termlineSpeed (&newLineState, 38400) < 0)
	{ fprintf (stderr, "panTiltOpen: %d not available\n");
	  return -1;
	}
	termlineSet (fd, &newLineState);

#if DEBUG_PANTILT_RATE
	if (panTiltCheckSysState (fd) < 0 )
	{ fprintf (stderr, "panTiltGetSysState: error requesting state\n");
	  return -1;
	}
	printf("Firmware fine!\n");
#endif
	if( panTiltSetBaud( fd, BAUD_38400 ) < 0 )
	{
	    int ibaud;
	    for( ibaud = 38400; ibaud >= 1200; ibaud /= 2 )
	    {
		if (termlineSpeed (&newLineState, ibaud) < 0)
		{ fprintf (stderr, "panTiltOpen: %d not available\n", ibaud);
		return -1;
		}
		termlineSet (fd, &newLineState);	
		if( panTiltSetBaud (fd, BAUD_38400) > 0 )
		{
		    fprintf( stderr, "old baud rate was %d\n", ibaud );
		    break;
		}
	    }
	}
	/* set final baud rate... */
	if (termlineSpeed (&newLineState, 38400) < 0)
	{ fprintf (stderr, "panTiltOpen: %d not available\n");
	  return -1;
	}
	termlineSet (fd, &newLineState);	
	if (panTiltSetBaud( fd, BAUD_38400 ) < 0 )
	{ fprintf (stderr, "panTiltOpen: can't set baud rate\n" );
	  return -1;
	}

	if (panTiltCheckSysState (fd) < 0 )
	{ fprintf (stderr, "panTiltGetSysState: error requesting state\n");
	  return -1;
	}
#if DEBUG_PANTILT_RATE
	printf("Firmware fine!\n");
#endif
	if (panTiltSetComSet (fd, SHORT_COMSET) < 0)
	 { fprintf (stderr, "panTiltOpen: error setting COMSET\n");
	   return -1;
	 }

	// Set the maximum equal the initial speed in order to avoid
	// S-curve moves!
	if (panTiltSetBaseCfg (fd, MAX_RATE, MAX_RATE, 16, 20) < 0)
	 { fprintf (stderr, "panTiltOpen: error setting base config\n");
	   return -1;
	 }
	if (panTiltSetHeadCfg (fd, MAX_RATE, MAX_RATE, 16, 20) < 0)
	 { fprintf (stderr, "panTiltOpen: error setting head config\n");
	   return -1;
	 }
	if (panTiltGetBaseCfg (fd, infoBuf) < 0 )
	 { fprintf (stderr, "panTiltOpen: error reading base config\n");
	 }
	else {
	  printf( "BaseCfg: max step: %d, start: %d, accel: %d, power: %d\n",
		  infoBuf[0],infoBuf[1],infoBuf[2],infoBuf[3]);
	}
	if (panTiltGetHeadCfg (fd, infoBuf) < 0 )
	 { fprintf (stderr, "panTiltOpen: error reading head config\n");
	 }
	else {
	  printf( "HeadCfg: max step: %d, start: %d, accel: %d, power: %d\n",
		  infoBuf[0],infoBuf[1],infoBuf[2],infoBuf[3]);
	}
#if 0
	if (panTiltSetCoordinates (fd, UPSIDE_DOWN) < 0)
	 { fprintf (stderr, "panTiltOpen: error setting COORDINATES\n");
	   return -1;
	 }
#endif
	return fd;
}

static int sendCom (int fd, int com, unsigned char *buf, int len, int noRep)
{
	aio_result_t res;
	int i;
	buf[0] = STX;
	buf[1] = noRep ? ADDR_1 : ADDR_0;
	buf[2] = com;
	buf[len-2] = computeCRC (buf, len-2);
	buf[len-1] = ETX;
#ifdef DEBUG
	printf ("wrte: ");
	for (i=0; i<len; i++)
	 { printf ("%2x ", buf[i]);
	 }
	printf ("\n");
#endif
tcflush(fd, TCIFLUSH);
	if (write (fd, buf, len) < 0)
	 { perror ("write");
	   return -1;
	 }
	return 1;
}

static int getReply (int fd, unsigned char *buf, int len)
{
	int n, i, k, ntry;
	aio_result_t res;
#if 0
tcflush(fd, TCIFLUSH);
#endif
	nap (0.50);
	n = 0;
	ntry = 0;
	do
	 { i = read( fd, &buf[n], len-n );
	   if (i < 0)
	   { perror ("read");
	     return -1;
	   }
#ifdef DEBUG
	   printf ("read %d: ", i);
	   for (k=0; k<i; k++)
	    { printf ("%2x ", (unsigned)buf[n+k]);
	    }
	   printf ("\n");
#endif
	   n += i;
	 }
	while (n < len);
	if (buf[2] != 0)
	 { fprintf (stderr, "getReply read error 0x%x\n", buf[2]);
	   return -1;
	 }
	if (buf[len-2] != computeCRC (buf, len-2))
	 { fprintf (stderr, "getReply read error, CRC=0x%x vs 0x%x\n",
			     buf[len-2], computeCRC (buf, len-2));
//	   return -1;
	 }
	return 1;
}

static int getReplyTimeout (int fd, unsigned char *buf, int len, int timeout)
{
	int n, i, k, ntry;
	aio_result_t res;
#if 0
tcflush(fd, TCIFLUSH);
#endif
	nap (0.50);
	n = 0;
	ntry = 0;
	do
	 { fd_set rd;
	   struct timeval timeval;
	   FD_ZERO(&rd);
	   FD_SET(fd,&rd);
	   timeval.tv_sec = timeout/1000000;
	   timeval.tv_usec = timeout%1000000;
	   i = select( fd+1, &rd, 0, 0, &timeval );
	   if( i < 0 )
	   { perror( "select" );
	     return -1;
	   }
	   if( i == 0 )
	   { printf( "getReply: timed out:read %d bytes\n", n );
	     printf ("read %d: ", n);
	     for (k=0; k<n; k++)
	     { printf ("%02x ", (unsigned)buf[n+k]);
	     }
	     printf ("\n");
	     return -1;
	   }
	   i = read( fd, &buf[n], len-n );
	   if (i < 0)
	   { perror ("read");
	     return -1;
	   }
#ifdef DEBUG
	   printf ("read %d: ", i);
	   for (k=0; k<i; k++)
	    { printf ("%2x ", (unsigned)buf[n+k]);
	    }
	   printf ("\n");
#endif
	   n += i;
	 }
	while (n < len);
	if (buf[2] != 0)
	 { fprintf (stderr, "getReply read error 0x%x\n", buf[2]);
	   return -1;
	 }
	if (buf[len-2] != computeCRC (buf, len-2))
	 { fprintf (stderr, "getReply read error, CRC=0x%x vs 0x%x\n",
			     buf[len-2], computeCRC (buf, len-2));
//	   return -1;
	 }
	return 1;
}

static int getReplyAsync (int fd, PanTiltReply *rep, int nd)
{
	int flags;
	int n = 0;

	if (PANTILT_REP_BUFSIZE-rep->cnt > 0)
	 { 
	   flags = fcntl (fd, F_GETFL);
	   fcntl (fd, F_SETFL, flags | O_NDELAY);
	   n = read (fd, rep->bp, PANTILT_REP_BUFSIZE-rep->cnt);
	   fcntl (fd, F_SETFL, flags);
	   if (n < 0)
	    { perror ("panTilt:getReplyAsync:read");
	      return -1;
	    }
	   else if (n > 0)
	    { rep->bp += n;
	      rep->cnt += n;
	    }
	 }
	if (rep->cnt > nd)
	 { fprintf (stderr, "panTilt:getReplyAsync: got too many bytes (%d)",
		    rep->cnt);
	   return -1;
	 }
	else if (rep->cnt == nd)
	 { return 1;
	 }
	else if (rep->cnt < nd)
	 { return 0;
	 }
}

static void packInt (char *buf, int ival)
{
	union spack pack;

	pack.s = (short)ival;
	buf[0] = pack.c[1];
	buf[1] = pack.c[0];
}

static int unpackInt (char *buf)
{ 
	union spack pack;

	pack.c[1] = buf[0];
	pack.c[0] = buf[1];
	return pack.s;
}

void panTiltClose (int fd)
{
	termlineSet (fd, &oldLineState);
	close (fd);
}

int panTiltSetPos (int fd, int *new, int *old, double interval)
{
	int del[2];
	char buf[64];
	int rate;

	packInt (&buf[3], new[0]);
	packInt (&buf[5], new[1]);
	rate = ABS ((new[0]-old[0])/interval);
	if (rate == 0)
	  rate = MAX_RATE;
	// printf ("Interval=%f\n", interval);
	// printf ("rate=%d\n", rate);
	packInt (&buf[7], rate);
	rate = ABS ((new[1]-old[1])/interval);
	if (rate == 0)
	  rate = MAX_RATE;
	// printf ("rate=%d\n", rate);
	packInt (&buf[9], rate);
	//exit(1);
	if (sendCom (fd, SET_POS_VEL, buf, 13, REPLY) < 0)
	 { fprintf (stderr, "panTiltSetPos: send error\n");
	   return -1;
	 }
	if (getReply (fd, buf, 5) < 0)
	 { fprintf (stderr, "panTiltSetPos: reply error\n");
	   return -1;
	 }
	return 1;
}

int panTiltSetPosDirect (int fd, int *new)
{
	int del[2];
	char buf[64];
	int rate = 4096;

	// printf ("In panTiltSetPosDirect: %d %d\n", new[0], new[1]);
	packInt (&buf[3], new[0]);
	packInt (&buf[5], new[1]);
	if (sendCom (fd, SET_POS, buf, 9, REPLY) < 0)
	 { fprintf (stderr, "panTiltSetPosDirect: send error\n");
	   return -1;
	 }
	if (getReply (fd, buf, 5) < 0)
	 { fprintf (stderr, "panTiltSetPosDirect: reply error\n");
	   return -1;
	 }
	return 1;
}

int panTiltSetPosAsync (int fd, int *new, int *old, double interval)
{
	int del[2];
	char buf[64];
	int rate;


	packInt (&buf[3], new[0]);
	packInt (&buf[5], new[1]);
#if DEBUG_PANTILT_RATE
	printf("Pos = %d  %d to %d %d \n", old[0], old[1], new[0], new[1] );
	printf("In Buffer = %x %x  - %x %x \n", 
	       buf[3], buf[4], buf[5], buf[6] );
#endif
	rate = ABS ((new[0]-old[0])/interval);
	if (rate == 0)
	  rate = MAX_RATE;
	if ( rate > MAX_RATE ) rate = MAX_RATE;
	packInt (&buf[7], rate);
#if DEBUG_PANTILT_RATE  
	printf ("Interval=%f\n", interval);
	printf ("rate=%d   ", rate);
	printf("In Buffer = %x %x  \n", buf[7], buf[8] );
#endif
	rate = ABS ((new[1]-old[1])/interval);
	if (rate == 0)
	  rate = MAX_RATE;
	if ( rate > MAX_RATE ) rate = MAX_RATE;
	packInt (&buf[9], rate);
#if DEBUG_PANTILT_RATE
	printf ("rate=%d  ", rate);
	printf("In Buffer = %x %x  \n", buf[9], buf[10] );
	//exit(1);
#endif
	if (sendCom (fd, SET_POS_VEL, buf, 13, REPLY) < 0)
	 { fprintf (stderr, "panTiltSetPos: send error\n");
	   return -1;
	 }
	return 1;
}

int panTiltSetPosReply (int fd, PanTiltReply *rep)
{
	int stat;
	
	stat = getReplyAsync (fd, rep, 5);
	if (stat < 0)
	 { fprintf (stderr, "panTiltSetPosReply: error in reply\n");
	   return -1;
	 }
	else
	 { return stat;
	 }
}

int panTiltSetComSet (int fd, int comset)
{
	char buf[64];

	buf[3] = comset;
	if (sendCom (fd, SET_COMSET, buf, 6, REPLY) < 0)
	 { fprintf (stderr, "panTiltSetComSet: send error\n");
	   return -1;
	 }
	if (getReply (fd, buf, 5) < 0)
	 { fprintf (stderr, "panTiltSetComSet: reply error\n");
	   return -1;
	 }
	return 1;				
}

int panTiltSetCoordinates (int fd, int coords)
{
	char buf[64];

	buf[3] = coords;
	if (sendCom (fd, SET_COORDINATES, buf, 6, REPLY) < 0)
	 { fprintf (stderr, "panTiltSetCoordinates: send error\n");
	   return -1;
	 }
	if (getReply (fd, buf, 6) < 0)
	 { fprintf (stderr, "panTiltSetCoordinates: reply error\n");
	   return -1;
	 }
	return 1;				
}

int panTiltGetCoordinates (int fd, int* coords)
{
	char buf[64];

	if (sendCom (fd, GET_COORDINATES, buf, 6, REPLY) < 0)
	 { fprintf (stderr, "panTiltGetCoordinates: send error\n");
	   return -1;
	 }
	if (getReply (fd, buf, 6) < 0)
	 { fprintf (stderr, "panTiltGetCoordinates: reply error\n");
	   return -1;
	 }
	*coords = buf[3];
	return 1;				
}

int panTiltSetAddress (int fd, int coords)
{
	char buf[64];

	buf[3] = ADDR_1;
	if (sendCom (fd, SET_ADDRESS, buf, 6, REPLY) < 0)
	 { fprintf (stderr, "panTiltSetAddress: send error\n");
	   return -1;
	 }
	return 1;				
}

int panTiltGetPos (int fd, int *pos)
{
	char buf[64];
		
	if (sendCom (fd, GET_POS, buf, 5, REPLY) < 0)
	 { fprintf (stderr, "panTiltGetPos: send error\n");
	   return -1;
	 }
	if (getReply (fd, buf, 9) < 0)
	 { fprintf (stderr, "panTiltGetPos: reply error\n");
	   return -1;
	 }
	pos[0] = unpackInt (&buf[3]);
	pos[1] = unpackInt (&buf[5]);
	return 1;
}

int panTiltGetPosAsync (int fd)
{
	char buf[64];

printf ("GPA\n");		
	if (sendCom (fd, GET_POS, buf, 5, REPLY) < 0)
	 { fprintf (stderr, "panTiltGetPos: send error\n");
	   return -1;
	 }
	return 1;
}

int panTiltGetPosReply (int fd, int *pos, PanTiltReply *rep)
{
	int stat;

	stat = getReplyAsync (fd, rep, 9);
	if (stat < 0)
	 { fprintf (stderr, "panTiltSetPosReply: error in reply\n");
	   return -1;
	 }
	else if (stat == 1)
	 { pos[0] = unpackInt (&rep->buf[3]);
	   pos[1] = unpackInt (&rep->buf[5]);
	   return 1;
	 }
	else
	 { return 0;
	 }
}

int panTiltSetBaud (int fd, int baud)
{
	char buf[64];

	buf[3] = baud;
	if (sendCom (fd, 0x0a, buf, 6, REPLY) < 0)
	 { fprintf (stderr, "panTiltSetBaud: send error\n");
	   return -1;
	 }
	if( sendCom( fd, 0x0D, buf, 5, REPLY ) < 0 )
	{
	    fprintf( stderr, "panTiltSetBaud: send query baud error\n" );
	    return -1;
	}
	if( getReplyTimeout( fd, buf, 9, 500000 ) < 0 )
	{
	    fprintf( stderr, "panTiltSetBaud: recv query baud error\n" );
	    return -1;
	}
	if( buf[3] != 'B' && buf[4] != 'A' && buf[5] != 'U' && buf[6] != 'D' )
	{
	    fprintf( stderr, "panTiltSetBaud: recv query baud bad\n" );
	    return -1;
	}

	return 1;				
}

int panTiltCheckBaud (int fd)
{
	char buf[64];

	if (sendCom (fd, CHECK_BAUD, buf, 5, REPLY) < 0)
	{ fprintf(stderr, "panTiltCheckBaud: send error\n");
	  return -1;
	}
	if ( getReply (fd, buf, 9) < 0)
	{ fprintf(stderr, "panTiltCheckBaud: reply error\n");
	  return -1;
	}
	if (strncmp( buf+3, "BAUD", 4) != 0)
	{ fprintf(stderr, "panTiltCheckBaud: auto-baud detection failed\n");
	  return -1;
	}
	return 1;
}

int panTiltCalibrate (int fd)
{
	char buf[64];
		
	if (sendCom (fd, CALIBRATE, buf, 5, REPLY) < 0)
	 { fprintf (stderr, "panTiltCalibrate: send error\n");
	   return -1;
	 }
	if (getReply (fd, buf, 9) < 0)
	 { fprintf (stderr, "panTiltCalibrate: reply error\n");
	   return -1;
	 }
	return 1;
}

int panTiltVerifyPos(int fd, int mode)
{
	char buf[64];
	int azimuth_err, elevation_err;

	buf[3] = mode;
	if (sendCom (fd, VERIFY_POS, buf, 6, REPLY) < 0)
	 { fprintf (stderr, "panTiltVerifyPos: send error\n");
	   return -1;
	 }
	if (getReply (fd, buf, 9) < 0)
	 { fprintf (stderr, "panTiltVerifyPos: reply error\n");
	   return -1;
	 }
	azimuth_err = unpackInt (&buf[3]);
	if (azimuth_err != 0)
	 { fprintf (stderr, "panTiltVerifyPos: azimuth no reference error\n");
	   return -1;
	 }
	elevation_err = unpackInt (&buf[5]);
	if (elevation_err != 0)
	 { fprintf (stderr, "panTiltVerifyPos: elevation no reference error\n");
	   return -1;
	 }
	return 1;				
}

int panTiltGetReferPos(int fd, int* AziExtA, int* EleExtA, int* AziRefA,
	int* EleRefA, int* AziRefB, int* EleRefB, int* AziExtB, int* EleExtB)
{
	char buf[64];

	if (sendCom (fd, GET_REFER_POS, buf, 5, REPLY) < 0)
	 { fprintf (stderr, "panTiltGetReferPos: send error\n");
	   return -1;
	 }
	if (getReply (fd, buf, 21) < 0)
	 { fprintf (stderr, "panTiltyGetReferPos: reply error\n");
	   return -1;
	 }
	*AziExtA = unpackInt (&buf[3]);
	*EleExtA = unpackInt (&buf[5]);
	*AziRefA = unpackInt (&buf[7]);
	*EleRefA = unpackInt (&buf[9]);
	*AziRefB = unpackInt (&buf[11]);
	*EleRefB = unpackInt (&buf[13]);
	*AziExtB = unpackInt (&buf[15]);
	*EleExtB = unpackInt (&buf[17]);
	return 1;				
}

int panTiltStop(int fd)
{
	char buf[64];

	if (sendCom (fd, STOP, buf, 5, REPLY) < 0)
	 { fprintf (stderr, "panTiltStop: send error\n");
	   return -1;
	 }
	if (getReply (fd, buf, 5) < 0)
	 { fprintf (stderr, "panTiltStop: reply error\n");
	   return -1;
	 }
	return 1;				
}

int panTiltSetBaseCfg (int fd, int maxrate, int startrate,
			       int accel, int powerTimeOut)
{
	char buf[64];

	packInt (&buf[3], maxrate);
	packInt (&buf[5], startrate);
	packInt (&buf[7], accel);
	packInt (&buf[9], powerTimeOut);
	if (sendCom (fd, SET_BASE_CFG, buf, 13, REPLY) < 0)
	 { fprintf (stderr, "panTiltSetBaseCfg: send error\n");
	   return -1;
	 }
	if (getReply (fd, buf, 5) < 0)
	 { fprintf (stderr, "panTiltStop: reply error\n");
	   return -1;
	 }
	return 1;				
}

int panTiltGetBaseCfg (int fd, int *info)
{
	char buf[64];

	if (sendCom (fd, GET_BASE_CFG, buf, 5, REPLY) < 0)
	 { fprintf (stderr, "panTiltGetBaseCfg: send error\n");
	   return -1;
	 }
	if (getReply (fd, buf, 13) < 0)
	 { fprintf (stderr, "panTiltGetBaseCfg: reply error\n");
	   return -1;
	 }
	info[0] = unpackInt (&buf[3]);
	info[1] = unpackInt (&buf[5]);
	info[2] = unpackInt (&buf[7]);
/*	info[3] = 0;*/
	info[3] = unpackInt (&buf[9]); 
	return 1;				
}

int panTiltGetHeadCfg (int fd, int *info)
{
	char buf[64];

	if (sendCom (fd, GET_HEAD_CFG, buf, 5, REPLY) < 0)
	 { fprintf (stderr, "panTiltGetHeadCfg: send error\n");
	   return -1;
	 }
	if (getReply (fd, buf, 13) < 0)
	 { fprintf (stderr, "panTiltGetHeadCfg: reply error\n");
	   return -1;
	 }
	info[0] = unpackInt (&buf[3]);
	info[1] = unpackInt (&buf[5]);
	info[2] = unpackInt (&buf[7]);
	/*	info[3] = 0; JL */
 	info[3] = unpackInt (&buf[9]);
	return 1;				
}

int panTiltSetHeadCfg (int fd, int maxrate, int startrate,
			       int accel, int powerTimeOut)
{
	char buf[64];

	packInt (&buf[3], maxrate);
	packInt (&buf[5], startrate);
	packInt (&buf[7], accel);
	packInt (&buf[9], powerTimeOut);
	if (sendCom (fd, SET_HEAD_CFG, buf, 13, REPLY) < 0)
	 { fprintf (stderr, "panTiltSetHeadCfg: send error\n");
	   return -1;
	 }
	if (getReply (fd, buf, 5) < 0)
	 { fprintf (stderr, "panTiltSetHeadCfg: reply error\n");
	   return -1;
	 }
	return 1;				
}

int panTiltCheckSysState ( int fd ) 
{
  char buf[16];
  if (sendCom (fd, GET_SYS_STATE, buf, 5, REPLY) < 0)
    { fprintf (stderr, "panTiltGetSysState: send error\n");
    return -1;
    }
  if (getReply (fd, buf, 6) < 0)
    { fprintf (stderr, "panTiltGetSysState: reply error\n");
    return -1;
    }
  if ( buf[2] != 0 )
    { fprintf (stderr, "panTiltGetSysState: error byte set\n");
    return -1;
    }
  if ( buf[3] != (char) 1 )
    { fprintf (stderr, "panTiltGetSysState: firmware not executing\n");
    return -1;
    }
  return 1;
}
#endif

