/*-=-=-=-=-=-=-=-=-=-=-=- RCCL/RCI Notice of Copyright -=-=-=-=-=-=-=-=-=-=-*
 *									    *
 *	Copyright (C) 1987, 1991 by John E. Lloyd.			    *
 *									    *
 *	Information on copyright permissions, as well as a complete	    *
 *	disclaimer of warranty, is given in the file COPYRIGHT.		    *
 *									    *
 *-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=*/

#ifndef _MUTEX_H_
#define _MUTEX_H_

#include "../h/protoDef.h"

#if defined(Sun4sol_TARGET) || defined(X86sol_TARGET)
#include <synch.h>
#endif

#ifdef __cplusplus
extern "C"
{
#endif

/* The magic number has been put in so that we can verify the structure
   is a real mutex and avoid infinite loops.
*/

#define RCI_MUTEX_MAGIC	0x87654321

#if !defined(Sun4sol_TARGET) && !defined(X86sol_TARGET)

/* This data structure is used to implement mutual exclusion among (up
   to four) independent processors, based on the generalized Decker's
   algorithm. */

typedef struct {			/* mutual exclusion flag */
	union {
	  unsigned long l;
	  unsigned char p[4]; } need;
	long check;
} MUTEX;

#define MUTEX_CLEAR(mx)	((mx)->need.l = 0, (mx)->check = RCI_MUTEX_MAGIC)
#define MUTEX_VALID(mx)	((mx)->check == RCI_MUTEX_MAGIC)

#else

typedef struct {
	mutex_t m;
	long check;
} MUTEX;

#define MUTEX_CLEAR(mx)	(mutex_init(&(mx)->m, USYNC_THREAD, 0), \
			 (mx)->check = RCI_MUTEX_MAGIC)

#define MUTEX_VALID(mx)	((mx)->check == RCI_MUTEX_MAGIC)

#endif /* !defined(Sun4sol_TARGET) && !defined(X86sol_TARGET) */

void mutexInitialize PROTO((int));
int mutexClear PROTO((MUTEX *));
int mutexBegin PROTO((MUTEX *));
int mutexEnd PROTO((MUTEX *));

#ifdef __cplusplus
}
#endif

#endif
