/*=-=-=-=-=-=-=-=-=-=-  RCCL/RCI Notice of Copyright  -=-=-=-=-=-=-=-=-=-=-=*
 *                                                                          *
 *      Copyright (C) 1987 by all authors described below. Permission to    *
 *      use, distribute and copy is granted, but only for non-commercial    *
 *      purposes and providing the authors are acknowledged and this        *
 *      copyright notice is displayed prominently.                          *
 *                                                                          *
 *      The authors provide this software in hope that it will be useful,   *
 *      but do not assume responsibility for its use.                       *
 *                                                                          *
 *=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-*/

#include <stdio.h>
#include <ctype.h>
#include <pwd.h>
#include <sys/wait.h>
#include <sys/param.h> /* for MAXPATHLEN */

#include "../h/stringDefs.h"
#include "../h/cdefs.h"
#include "../h/protoDef.h"

#define NEW_RCC_LOCATE  1       /* new way of locating .rcc files */

#ifdef HAS_UNION_WAIT
#define WAIT_T union wait
#else
#define WAIT_T int
#endif

char **argvec;
int argcnt;

char *malloc();
char *strcpy();
char *getenv();
char *expandFileName();

#define MAXARGS 256

#define YES     1
#define NO      0

typedef struct _arg {
	struct _arg *next;
	char *val;
	char *printVal;
	int flags;
} ArgDesc;

typedef struct {
	ArgDesc *head;
	ArgDesc *tail;
} ArgList;

ArgDesc *addArg PROTO((ArgList*, char*));
ArgDesc *addArgList PROTO((ArgList*, ArgList*));
void clearList PROTO((ArgList*));
int argListSize PROTO((ArgList*));
char *readline PROTO((char*, int, FILE*));
void readcf PROTO((char*));
void config PROTO(());
char *suffix PROTO((const char*));
int isLibraryArg PROTO((int, char**));
int isSourceModule PROTO((const char*));
int isObjectModule PROTO((const char*));
char *setsuf PROTO((char*, char*));
void argNumCheck PROTO((int, char**, int));
void do_command PROTO((char*, ArgList*));
char *expandFileName PROTO((char*, char*));

ArgDesc *addArg (list, val)
ArgList *list;
char *val;
{
	ArgDesc *new;

	new = (ArgDesc*)malloc(sizeof(ArgDesc));
	new->val = val;
	new->printVal = val;
	new->flags = 0;
	new->next = NULL;
	if (list->head == NULL)
	 { list->head = new;
	 }
	else
	 { list->tail->next = new;
	 }
	list->tail = new;
	return (new);
}

ArgDesc *addArgList (list, newargs)
ArgList *list, *newargs;
{
	ArgDesc *desc;
	ArgDesc *new = NULL;
	
	for (desc=newargs->head; desc; desc=desc->next)
	 { new = addArg (list, desc->val);
	   new->flags = desc->flags;
	   new->printVal = desc->printVal;
	 }
	return (new);	
}

void clearList (list)
ArgList *list;
{
	list->head = list->tail = NULL;
}

argListSize (list)
ArgList *list;
{
	ArgDesc *desc;
	int size = 0;

	for (desc=list->head; desc; desc=desc->next)
	 { size++;
	 }
	return (size);
}

#if NEW_RCC_LOCATE == 0
char *RCCL_env;
#endif
char *HOME_env;
struct files {
	char *keyword;
	char *format;
	int extra;
	ArgList *args;
  } files[] = { { "begin", "%s", 0 },
#define BEGIN (&files[0])
		{ "end", "%s", 0 },
#define END (&files[1])
		{ "ctrl", "%s", 0 },
#define CTRL (&files[2])
		{ "util", "%s", 0 },
#define UTIL (&files[3])
		{ "user", "%s", 0 },
#define USER (&files[4])
		{ "include", "-I%s", 2 },
#define INCLUDE (&files[5])
		{ "cc", "%s", 0 },
#define CC_CMD (&files[6])
		{ "libpath", "%s", 0 },
#define LIBPATH (&files[7])
		};

char *configfile = ".rcc";
char *target = NULL;
#define IS_TARGET(x)	(strcmp(target,x)==0)
#ifdef _AIX
char *default_cc = "bsdcc";
#else
char *default_cc = "cc";
#endif
char *C_command;

char output[80];

char *ofile;
int cflag;
int debuginfo;

char *readline (buf, size, fp)
char *buf;
int size;
FILE *fp;
{
	return (fgets (buf, size, fp));
}

void readcf(cfn)
char *cfn;
{
	FILE *cf;
	char *cp;
	char *key;
	char *kend;
	char line[0x10000];
	int i;

	cf = fopen(cfn, "r");
	if (cf == 0)
	 { return;
	 }
	while (readline(line,sizeof(line),cf) == line)
	 { cp = &line[0];
	   for (;isascii(*cp)&&isspace(*cp);cp++) ;
	   key = cp;
	   if (!*key || (*key == '#'))
	    { continue;
	    }
	   for (;isascii(*cp)&&isalnum(*cp);cp++) ;
	   kend = cp;
	   for (;isascii(*cp)&&isspace(*cp);cp++) ;
	   if (*cp != '=')
	    { fprintf(stderr,"%s: config file %s: missing or misplaced =\n",
			       argvec[0],cfn);
	      continue;
	    }
	   cp ++;
	   *kend = '\0';
	   for (i=0;i<(sizeof(files)/sizeof(files[0]));i++)
	    { if (! strcmp(key,files[i].keyword))
	       { break;
	       }
	    }
	   if (i >= sizeof(files)/sizeof(files[0]))
	    { fprintf(stderr,"%s: config file %s: unrecognized keyword %s\n",
		       argvec[0],cfn,key);
	      continue;
	    }
	   else
	    { char *base;
	      clearList (files[i].args);

	      while (1)
	       { ArgDesc *desc;
		 char *expand;

		 for (;isascii(*cp)&&isspace(*cp);cp++); /* skip whitespace */
		 if (!*cp)
		  { break;
		  }
		 base = cp;
		 for (;isascii(*cp)&&!isspace(*cp);cp++);

		 if (*cp)
		  { *cp++ = '\0';
		  }
		 expand = expandFileName(base, cfn);
		 desc = addArg (files[i].args, NULL);
		 desc->val = malloc(strlen(expand)+files[i].extra+1);
		 sprintf(desc->val,files[i].format,expand);
		 desc->printVal = malloc(strlen(base)+files[i].extra+1);
		 sprintf(desc->printVal,files[i].format,base);
	       }
	    }
	 }
	fclose(cf);
}

void config()
{
	char cf[MAXPATHLEN];
	int i;

	for (i=0;i<(sizeof(files)/sizeof(files[0]));i++)
	 { files[i].args = (ArgList*)malloc(sizeof(ArgList));
	   clearList (files[i].args);
	 }
#if NEW_RCC_LOCATE == 0
	RCCL_env = getenv("RCCL");
	if (RCCL_env == 0)
	 { struct passwd *pwd;
	   fprintf(stderr,"Can't use rcc without RCCL environment variable\n");
	   pwd = getpwnam("rccl");
	   if (pwd)
	    { fprintf(stderr,"Please add the following line to your .cshrc:\n");
	      fprintf(stderr,"\tsetenv RCCL ~rccl\n");
	    }
	   exit(1);
	 }
	sprintf(cf,"%s/conf/%s",RCCL_env,configfile);
	readcf(cf);
#else
	if (getenv("RCCL_CONF") == 0)
	 { struct passwd *pwd;
	   fprintf(stderr,"Can't use rcc without RCCL_CONF environment variable\n");
	   exit(1);
	 }
	if (target == NULL)
	 { char *cp;
	   if (getenv("RCCL_TARGETS") == 0)
	    { fprintf(stderr, "\
Target must be specified, using either the TARGET command option,\n\
or as the first word of the RCCL_TARGETS environment variable\n");
	      exit (1);
	    }
	   target = malloc(strlen(getenv("RCCL_TARGETS"))+1);
	   if (sscanf (getenv("RCCL_TARGETS"), "%s", target) == 0)
	    { fprintf(stderr, "\
Environment variable RCCL_TARGETS contains only whitespace\n");
	      exit (1);
	    }
	 }
	sprintf(cf,"%s/%s/%s",getenv("RCCL_CONF"),target,configfile);
	readcf(cf);
#endif
	HOME_env = getenv("HOME");
	if (HOME_env == 0)
	 { fprintf(stderr,"Don't know where your home directory is;\n");
	   fprintf(stderr,"  won't see any %s file there\n",configfile);
	 }
	else
	 { sprintf(cf,"%s/%s",HOME_env,configfile);
	   readcf(cf);
	 }
	readcf(configfile);
}


char *suffix(fn)
const char *fn;
{
	char *cp;

	cp = charRindex(fn,'.');
	return (cp ? cp : "");
}

int isLibraryArg (argc, argv)
int argc;
char **argv;
{
	if (ARGUMENT("-L") || ARGUMENT("-l"))
	 { argNumCheck (argc, argv, 2);
	   return (2);
	 }
	else if (ARGUMENTS("-L") || ARGUMENTS("-l"))
	 { return (1);
	 }
	else if (strcmp(suffix(*argv), ".a") == 0)
	 { return (1);
	 }
	else
	 { return (0);
	 }
}

int isSourceModule (arg)
const char *arg;
{
	if (strcmp (suffix(arg), ".c") == 0 ||
	    strcmp (suffix(arg), ".s") == 0)
	 { return (1);
	 }
	else
	 { return (0);
	 }
}

int isObjectModule (arg)
const char *arg;
{
	if (strcmp (suffix(arg), ".o") == 0)
	 { return (1);
	 }
	else
	 { return (0);
	 }
}
	
char *setsuf(fn,suf)
char *fn;
char *suf;
{
	char *sp;
	char *cp;

	sp = charRindex(fn,'.');
	if (! sp)
	 { int i = strlen(fn);
	   cp = malloc(i+strlen(suf)+1);
	   strcpy(cp,fn);
	   strcpy(cp+i,suf);
	   return(cp);
	 }
	cp = malloc((sp-fn)+strlen(suf)+1);
	BCOPY(fn,cp,sp-fn);
	strcpy(cp+(sp-fn),suf);
	return(cp);
}

int setDefaultOfile = 0;
int ignoreCtrl = 1;
ArgList planLibs;
ArgList ctrlLibs;
ArgList planArgs;
ArgList ctrlArgs;
ArgList cmdArgs;

int inCtrlPart = 0;

void argNumCheck(argc, argv, n)
int argc;
char **argv;
int n;
{
	if (argc < n)
	 { if (n-argc == 1)
	    { fprintf (stderr, "%s: %s needs another argument\n", 
				argvec[0], *argv);
	    }
	   else
	    { fprintf (stderr, "%s: %s needs %d more arguments\n", 
				argvec[0], *argv, (n-argc));
	    }
	   exit (1);
	 }
}

main (argc, argv)
int argc;
char **argv;
{
	int num;

	clearList (&planLibs);
	clearList (&ctrlLibs);
	clearList (&planArgs);
	clearList (&ctrlArgs);
	
	ofile = 0;
	cflag = 0;
	debuginfo = 0;

	/* make a first pass through the arguments to pull out
	   options controlling rcc */

	argcnt = argc;
	argvec = argv;
	argc--; argv++;
	while (argc)
	 { if (ARGUMENT("LIBFILE"))
	    { argNumCheck (argc, argv, 2);
	      configfile = argv[1];
	      argc-=2; argv+=2;
	    }
	   else if (ARGUMENT("TARGET"))
	    { argNumCheck (argc, argv, 2);
	      target = argv[1];
	      argc-=2; argv+=2;
	    }
	   else if (ARGUMENT("DEBUG"))
	    { debuginfo ++;
	      argc--; argv++;
	    }
	   else
	    { argc--; argv++;
	    }
	 }
	config();

	if (IS_TARGET("UVax") || IS_TARGET("Sun4"))
	 { setDefaultOfile = 1;
	   ignoreCtrl = 0;
	 }

	/* now make a second pass through the arugments to pull out
	   the arguments to be passed on to the cc command */

	argc = argcnt;
	argv = argvec;
	argc--; argv++;
	while (argc)
	 { if (ARGUMENT("CTRL"))
	    { if (!ignoreCtrl)
	       { if (inCtrlPart)
		  { fprintf(stderr, 
"%s: CTRL already specified, ignoring extra copy\n", argvec[0]);
		  }
		 else
		  { inCtrlPart = 1;
		  }
	       }
	      argc--; argv++;
	    }
	   else if (ARGUMENT("LIBFILE") ||
		    ARGUMENT("TARGET"))
	    { argc-=2; argv+=2;
	    }
	   else if (ARGUMENT("DEBUG"))
	    { argc--; argv++;
	    }
	   else if (ARGUMENT("-o"))
	    { argNumCheck (argc, argv, 2);
	      ofile = argv[1];
	      argc-=2; argv+=2;
	    }
	   else if (!ignoreCtrl && (num = isLibraryArg(argc, argv)))
	    { while (num--)
	       { addArg (inCtrlPart ? &ctrlLibs : &planLibs, *argv);
		 argc--; *argv++;
	       }
	    }
	   else if (ARGUMENT("-c"))
	    { cflag = 1;
	      argc--; argv++;
	    }
	   else
	    { addArg (inCtrlPart ? &ctrlArgs : &planArgs, *argv);
	      argc--; *argv++;
	    }
	 }

	if (!ofile && setDefaultOfile)
	 { ArgDesc *desc;

	   for (desc=planArgs.head; desc; desc=desc->next)
	    { if (isSourceModule(desc->val) || isObjectModule(desc->val))
	       { break;
	       }
	    }
	   if (!desc)
	    { for (desc=ctrlArgs.head; desc; desc=desc->next)
	       { if (isSourceModule(desc->val) || isObjectModule(desc->val))
		  { break;
		  }
	       }
	    }
	   if (desc)
	    { char *s = charRindex(desc->val,'/');
	      ofile = setsuf(s?s+1:desc->val,"");
	    }
	 }
	if (debuginfo)
	 { ArgDesc *desc;
	   int i;

	   printf("%d plan libraries:\n", argListSize(&planLibs));
	   for (desc=planLibs.head; desc; desc=desc->next)
	    { printf("\t%s\n", desc->val);
	    }
	   printf("%d ctrl libraries:\n", argListSize(&ctrlLibs));
	   for (desc=ctrlLibs.head; desc; desc=desc->next)
	    { printf("\t%s\n", desc->val);
	    }
	   printf("%d plan args:\n", argListSize(&planArgs));
	   for (desc=planArgs.head; desc; desc=desc->next)
	    { printf("\t%s\n", desc->val);
	    }
	   printf("%d ctrl args:\n", argListSize(&ctrlArgs));
	   for (desc=ctrlArgs.head; desc; desc=desc->next)
	    { printf("\t%s\n", desc->val);
	    }
	   if (ofile)
	    { printf("-o %s\n",ofile);
	    }
	   for (i=0;i<(sizeof(files)/sizeof(files[0]));i++)
	    { 
	      printf("%s=",files[i].keyword);
	      for (desc=files[i].args->head; desc; desc=desc->next)
	       { if (desc != files[i].args->head)
		  { printf(" ");
		  }
		 printf("%s",desc->printVal);
	       }
	      printf("\n");
	    }
	 }
	if (argListSize(CC_CMD->args) == 0)
	 { addArg (CC_CMD->args, default_cc);
	 }
        doPreCompile();
	if (cflag)
	 { doCflagCompile();
	 }
	else
	 { doGeneralCompile();
	 }
	exit(0);
}

doPreCompile()
{
	ArgDesc *desc;
	int numSourceMods = 0;

	clearList (&cmdArgs);

	addArgList (&cmdArgs, CC_CMD->args);
	addArgList (&cmdArgs, INCLUDE->args);
	addArg (&cmdArgs, "-c");

	if (IS_TARGET("UVax"))
	 { addArg (&cmdArgs, "-b");
	 }
	for (desc=planArgs.head; desc; desc=desc->next)
	 { if (!isSourceModule(desc->val) && !isObjectModule(desc->val))
	    { addArg (&cmdArgs, desc->val);
	    }
	 }
	for (desc=ctrlArgs.head; desc; desc=desc->next)
	 { if (!isObjectModule(desc->val))
	    { addArg (&cmdArgs, desc->val);
	      if (isSourceModule(desc->val))
	       { desc->val = setsuf(desc->val,".o");
		 numSourceMods++;
	       }
	    }
	 }
	if (numSourceMods > 0)
	 { do_command(cmdArgs.head->val, &cmdArgs);
	 }
}

doCflagCompile()
{
	ArgDesc *desc;
	int numSourceMods;

	clearList (&cmdArgs);

	addArgList (&cmdArgs, CC_CMD->args);
	addArgList (&cmdArgs, INCLUDE->args);
	addArg (&cmdArgs, "-c");

	if (!ignoreCtrl)
	 { /* control mods are already compiled; only do plan mods */

	   for (desc=planArgs.head; desc; desc=desc->next)
	    { if (!isObjectModule(desc->val))
	       { addArg (&cmdArgs, desc->val);
		 if (isSourceModule(desc->val))
		  { desc->val = setsuf(desc->val,".o");
		    numSourceMods++;
		  }
	       }
	    }
	   for (desc=ctrlArgs.head; desc; desc=desc->next)
	    { if (!isSourceModule(desc->val) && !isObjectModule(desc->val))
	       { addArg (&cmdArgs, desc->val);
	       }
	    }
	   if (numSourceMods > 0)
	    { do_command(cmdArgs.head->val, &cmdArgs);
	    }
	 }
	else
	 { /* just spit out everything */

	   for (desc=planArgs.head; desc; desc=desc->next)
	    { addArg (&cmdArgs, desc->val);
	    }
	   /* and there won't be any ctrl args ... */
	   do_command(cmdArgs.head->val, &cmdArgs);
	 }
}

doGeneralCompile()
{
	ArgDesc *desc;

	clearList (&cmdArgs);

	addArgList (&cmdArgs, CC_CMD->args);
	addArgList (&cmdArgs, INCLUDE->args);
	if (ofile)
	 { addArg (&cmdArgs, "-o");
	   addArg (&cmdArgs, ofile);
	 }
	for (desc=planArgs.head; desc; desc=desc->next)
	 { addArg (&cmdArgs, desc->val);
	 }
	for (desc=planLibs.head; desc; desc=desc->next)
	 { addArg (&cmdArgs, desc->val);
	 }
	addArgList (&cmdArgs, LIBPATH->args);
	addArgList (&cmdArgs, USER->args);

	if (IS_TARGET("UVax"))
	 { addArg (&cmdArgs, "-aa");
	 }
	if (IS_TARGET("Sun4"))
	 { addArg (&cmdArgs, "-Bstatic");
	 }
	if (!ignoreCtrl)
	 { addArgList (&cmdArgs, BEGIN->args);
	 }
	for (desc=ctrlArgs.head; desc; desc=desc->next)
	 { addArg (&cmdArgs, desc->val);
	 }
	addArgList (&cmdArgs, CTRL->args);
	addArgList (&cmdArgs, UTIL->args);
	for (desc=ctrlLibs.head; desc; desc=desc->next)
	 { addArg (&cmdArgs, desc->val);
	 }
	if (IS_TARGET("UVax"))
	 { addArg (&cmdArgs, "-aa");
	 }
	if (!ignoreCtrl)
	 { addArgList (&cmdArgs, END->args);
	 }
	do_command(cmdArgs.head->val, &cmdArgs);
}

void do_command(pgm, list)
char *pgm;
ArgList *list;
{
	ArgDesc *desc;
	char *args[MAXARGS];
	char **cpp;
	int num;
	int pid;

	num = 0;
	for (desc=list->head; desc; desc=desc->next)
	 { if (num >= MAXARGS-1)
	    { fprintf (stderr, "%s: too many arguments\n", argvec[0]);
	      exit (1);
	    }
	   args[num++] = desc->val;
	 }
	args[num] = NULL;

	printf("(%s)",pgm);
	for (desc=list->head; desc; desc=desc->next)
	 { printf (" %s", desc->printVal);
	 }
	putchar('\n');

	pid = fork();
	if (pid < 0)
	 { fprintf(stderr,"%s: cannot fork: ",argvec[0]);
	   perror(0);
	   exit(1);
	 }
	if (pid == 0)
	 { execvp(pgm,args);
	   fprintf(stderr,"%s: cannot exec %s: ",argvec[0],pgm);
	   perror(0);
	   exit(1);
	 }
	while (1)
	 { int p;
	   WAIT_T status;
	   p = wait(&status);
	   if (p < 0)
	    { break;
	    }
	   if (p != pid)
	    { fprintf(stderr,"%s: unexpected child %d died\n",argvec[0],p);
	    }
	   else
	    { if (WIFSIGNALED(status))
	       { static char *sigmsgs[] = { "Signal 0",
					    "Hanged up",
					    "Interrupted",
					    "Quit",
					    "Illegal instruction",
					    "Trace trap",
					    "IOT trap",
					    "EMT trap",
					    "Floating exception",
					    "Killed",
					    "Bus error",
					    "Segmentation violation",
					    "Bad syscall",
					    "Broken pipe",
					    "Alarm",
					    "Terminated",
					    "I/O condition",
					    "Stopped (signal)",
					    "Stopped",
					    "Continued",
					    "Child died",
					    "Tty input",
					    "Tty output",
					    "I/O possible",
					    "CPU limit exceeded",
					    "File size limit exceeded",
					    "Virtual alarm",
					    "Profiling alarm",
					    "Signal 28",
					    "Signal 29",
					    "Signal 30",
					    "Signal 31",
					    "Signal 32" };
		 fprintf(stderr,"%s",sigmsgs[WTERMSIG(status)]);
#ifdef HAS_UNION_WAIT
		 if (status.w_coredump)
		  { fprintf(stderr," (core dumped)");
		  }
#else
#ifdef WCOREFLG
		 if (status & WCOREFLG)
		  { fprintf(stderr," (core dumped)");
		  }
#endif
#endif
		 fprintf(stderr,"\n");
		 exit(1);
	       }
	      else if (WEXITSTATUS(status))
	       { fprintf(stderr,"Exit %d\n",WEXITSTATUS(status));
		 exit(WEXITSTATUS(status));
	       }
	      break;
	    }
	 }
}

char *expandFileName(name, cfn)
char *name;
char *cfn;
{
	       /* expands a file name. Assumes that the file name contains no
		  white space. Expands leading ~'s, plus environment
		  variables
		*/
	       
	       static char buf[MAXPATHLEN+64];
	       register char *cp, *np;
	       char *base;
	       char *tildedir;
	       char *tildetemp;
	       struct passwd *pwd;

	       cp = name;
	       np = buf;
	       *np = '\0';
	       
	       /* If the first character is ~, do home directory expansion */

	       if (*cp == '~')
		{ int tdlen;
		  base = cp;
		  for (cp = base+1; *cp && (*cp != '/'); cp++)
		     ;
		  if (cp == base+1)
		   { if ((tildedir = getenv("HOME")) == NULL)
		      { fprintf (stderr, 
"Config file %s: bare ~ used without HOME environment variable\n", cfn);
			exit (-1);
		      }
		   }
		  else
		   { char savec;
		     savec = *cp;
		     *cp = '\0';
		     if ((pwd = getpwnam(base+1)) == NULL)
		      { fprintf (stderr,
"Config file %s: no such user '%s'\n", cfn, base+1);
			exit (-1);
		      }
		     tildedir = pwd->pw_dir;
		     *cp = savec;
		   }
		  strcpy (np, tildedir);
		  np += strlen (np);
		}

	       /* Now proceed with regular expansion */

	       while (*cp)
		{ if (*cp == '\\')
		   { cp++;
		     if (*cp)
		      { *np++ = *cp++;
		      }
		   }
		  else if (*cp == '$')
		   { char savec;
		     char *env;

		     cp++;
		     base = cp;
		     for (; *cp && (isalnum(*cp) || *cp == '_'); cp++)
			;
		     savec = *cp;
		     *cp = '\0';
		     if (cp == base)
		      { fprintf (stderr, 
"Conf file %s: NULL environment variable in file name\n", cfn);
			exit (-1);
		      }
		     if ((env = getenv(base)) == NULL)
		      { fprintf (stderr, 
"Conf file %s: environment variable %s not defined\n", cfn, base);
			exit (-1);
		      }
		     strcpy (np, env);
		     np += strlen(np);
		     *cp = savec;
		   }
		  else
		   { *np++ = *cp++;
		   }
		}
	       *np = '\0';
	       return (buf);
}

