#include "../h/cdefs.h"
#include "../h/IOdevice.h"
#include "../gantry/mc8prim.h"
#include "linear_io.h"
#include "../h/pmc_pid.h"
#include "../h/moper_com.h"
#include "../h/errorCodes.h"
#include "../h/errorStack.h"
#include <sys/types.h>
#include <fcntl.h>

#include "pmcioctl.h"

#define PID_HEADER_BASE	0x25000

#define LINEAR_CHAN0	0

void static doclose (IO_DEVICE *dev)
{
	int k;

	if (dev->args[k] != -1)
	 { close (dev->args[k]);
	   dev->args[k] = -1;
	 }
}

int linear_attach (IO_DEVICE *dev)
{
	char exopenDevName[256];
	char *fxnName = "linear_attach()";
	char *cp;
	int fd[3];
	mc8_struct *pmc8Info;

	/* open the exclusive-open devices for each of the three
	   channels */
	strcpy (exopenDevName, dev->addrStr);
	strcat (exopenDevName, "chanx");
	cp = exopenDevName+strlen(exopenDevName)-1;

	dev->args[0] = -1;

	*cp = LINEAR_CHAN0 + '0';
	if ((fd[0] = open (exopenDevName, O_RDWR)) == -1)
	 { sysErrorMsg (EConnect, fxnName, "open chan 0");
	   return -1;
	 }

	/* Do the ioctl here to enable user mode access to I/O space.
	   We can use any of the open file descriptors. */
	if (ioctl (fd[0], PMCIOC_ENABLE_IOPORTS, 1) == -1 ||
	    ioctl (fd[0], PMCIOC_ENABLE_IOPORTS, 0) == -1)
	 { sysErrorMsg (EConnect, fxnName, "ioctl PMCIOC_ENABLE_IOPORTS");
	   doclose (dev);
	   return -1;
	 }

	pmc8Info = mc8Open(dev->address, 0, 0);
	if (pmc8Info == NULL)
	 { errorMsg (EConnect, fxnName, "mc8Open: no response from PMC\n");
	   doclose (dev);
	   return -1;
	 }
	dev->location = (char*)pmc8Info;	
	return 1;
}

void linear_detach (IO_DEVICE *dev)
{
	mc8Close ((mc8_struct *)dev->location);
	doclose (dev);
}

void linear_set_rate (IO_DEVICE *dev, int rate)
{
}

void linear_set_power (IO_DEVICE *dev, int power)
{
}

void linear_set_stop (IO_DEVICE *dev, int j)
{
}

void linear_set_encvalue (IO_DEVICE *dev, int val, int j)
{
	mc8_struct *pmc8 = (mc8_struct*)dev->location;
	
	mc8WriteMem (pmc8, PID_HEADER_OFF+PID_POSN_ADDR*j, (short*)&val, 2);
}

void linear_set_enc (IO_DEVICE *dev, int val, int j)
{
	mc8_struct *pmc8 = (mc8_struct*)dev->location;
	
	mc8WriteMem (pmc8, PID_HEADER_OFF+PID_GOAL_ADDR*j, (short*)&val, 2);
}

void linear_set_cur (IO_DEVICE *dev, int cur, int j)
{
}

void linear_set_integration (IO_DEVICE *dev, int mask)
{
}

void linear_set_calibMode (IO_DEVICE *dev, int mask)
{
}

int linear_get_state(IO_DEVICE *dev)
{
	return IS_CALIBRATED | POWER_ON;	
}

void linear_get_enc (IO_DEVICE *dev, int *enc)
{
	mc8_struct *pmc8 = (mc8_struct*)dev->location;

	mc8ReadMem (pmc8, PID_HEADER_OFF+PID_POSN_ADDR, (short*)enc, 2);
}


void linear_get_cur (IO_DEVICE *dev, int *rawTor)
{
	rawTor[0] = 0;
}

int linear_get_integration (IO_DEVICE *dev)
{
	return 0;
}

int linear_get_calibMode (IO_DEVICE *dev)
{
	return 0;
}

int linear_get_mark (IO_DEVICE *dev)
{
	return 0;
}

int linear_get_options (IO_DEVICE *dev)
{
	return 0;
}
