/*-=-=-=-=-=-=-=-=-=-=-=- RCCL/RCI Notice of Copyright -=-=-=-=-=-=-=-=-=-=-*
 *									    *
 *	Copyright (C) 1987, 1991 by John E. Lloyd.			    *
 *									    *
 *	Information on copyright permissions, as well as a complete	    *
 *	disclaimer of warranty, is given in the file COPYRIGHT.		    *
 *									    *
 *-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=*/

#include <stdio.h>
#include <signal.h>

#ifdef Vxw_TARGET
#define main playback_
#endif

#include "../h/stringDefs.h"
#include "../h/cdefs.h"
#include "../h/rciTask.h"
#include "../h/rciMacros.h"
#include "../h/robotMath.h"
#include "../h/rciOptions.h"
#include "../h/fatalMsg.h"
#include "macros.h"
#include "playback.h"

int numJoints;

RCI_DESC *td;
RCI_RBT *rbt;
HOW *how;
JLS *jls;

char *getDefaultRobot();

extern int tgenStart();
extern int tgenDrive();
extern int tgenRelease();

char **argvec;

main(argc, argv)
int argc;
char **argv;
{
	int j;
	char *robotName = NULL;
	char *fileName = NULL;
	char *simulatorAddr = NULL;
	FILE *fp;
	int simflag = NO;
	int turnPowerOff = YES;
	int interval = 20;	/* 20 msec */
	int status;
	TgenBuf *tbuf;
	double *bufp;
	unsigned long checking;

	checking = CHK_RBTALL;

	argvec = argv;

	argc--; argv++;
	while (argc)
	 {
	   if (ARGUMENTS ("-sim="))
	    { char *sname;
	      if (strlen(sname=charIndex(*argv,'=')+1) == 0)
	       { argc = -1;
		 break;
	       }
	      simulatorAddr = sname;
	      simflag = YES;
	    }
	   else if (ARGUMENT ("-sim"))
	    { simflag = YES;
	    }
	   else if (ARGUMENT ("-vcheckOff"))
	    { checking &= ~(CHK_REQVEL|CHK_MAXVEL);
	    }
	   else if (ARGUMENT("-on"))
	    { turnPowerOff = NO;
	    }	   
	   else if (ARGUMENT("-interval"))
	    { argc--; argv++;
	      if (argc == 0)
	       { argc = -1;
		 break;
	       }
	      interval = atoi(*argv);
	    }	   
	   else if (ARGUMENT("-f"))
	    { argc--; argv++;
	      if (argc == 0)
	       { argc = -1;
		 break;
	       }
	      fileName = *argv;
	    }	   
	   else if (!robotName)
	    { robotName = *argv;
	    }
	   else
	    { argc = -1;
	      break;
	    }
	   argc--; argv++;
	 }
	if (argc < 0 || fileName == NULL)
	 { usageMsg (-1, 
"%s -f <fileName> [<robotName>] [-sim[=<simulatorAddr>]] [-on] \n\
	[-interval <msec>]\n", 
	   *argvec);
	 }
	if (fileName)
	 { if ((fp = fopen (fileName, "r")) == NULL)
	    { perror ("fopen");
	      fprintf (stderr, "Can't open file %s\n", fileName);
	      exit (1);
	    }
	 }

	rciSetOptions (EXIT_ON_ERROR);
	if (!robotName)
	 { if ((robotName = getDefaultRobot()) == NULL)
	    { printErrors();
	      exit (-1);
	    }
	 }	

	if ((td = rciCreate ("tgen", /*cpus=*/0x1)) == NULL)
	 { printErrors();
	   exit (-1);
	 }
	if (simflag)
	 { if (simAttachRobot (td, robotName, simulatorAddr) < 0)
	    { printErrors();
	      exit (-1);
	    }
	   simScheduling (td);
	 }
	else
	 { if (rciAttachRobot (td, robotName, NULL) < 0)
	    { printErrors();
	      exit (-1);
	    }
	   if (rciScheduling (td, ON_TRIGGER, TASKROBOT(td,0)->idev) < 0)
	    { printErrors();
	      exit (-1);
	    }
	 }
	if (rciRate (td, interval, 0) < 0)
	 { printErrors();
	   exit(-1);
	 }
	if (rciControlFxns (td, tgenStart, tgenDrive, tgenRelease, 0)<0)
	 { printErrors();
	   exit (-1);
	 }

	tbuf = (TgenBuf*)rciSharedMemory ("tbuf", sizeof(TgenBuf), td);
	if (tbuf == NULL)
	 { printErrors();
	   exit (1);
	 }

	rbt = TASKROBOT(td, 0);
	how = rbt->how;
	jls = rbt->jls;
	numJoints = jls->numJ;

	/* read everything into memory */

	tbuf->go = 0;
	tbuf->idx = 0;
	bufp = tbuf->buf;
	do
	 { status = fscanVd (fp, bufp, numJoints);
	   if (status == 0)
	    { fprintf (stderr, "Error reading input, near word %d\n", 
				bufp-tbuf->buf);
	      exit (1);
	    }
	   else if (status == 1)
	    { bufp += numJoints;
	    }
	 }
	while (status == 1 && bufp-tbuf->buf <= TBUF_SIZE-numJoints);

	printf ("%d samples read\n", (bufp-tbuf->buf)/numJoints);
	if (status == 1)
	 { printf ("Warning: input may have been truncated\n");
	 }
	SET_CHECKING (rbt, checking);

	printf ("Hit <CR> when you are ready to move the robot\n");
	 { int c;
	   while ((c = getchar()) != '\n')
	    { printf ("c=%d\n", c);
	    }
	 }

	if (rciStart (td->taskCode) < 0)
	 { RCI_EXIT (-1);
	 }
	if (!turnPowerOff)
	 { SET_POWER_ON_RELEASE (rbt, 1) ;
	 }

	SET_POWER_ON(rbt);
	rciBlock(td);	
	rciBlock(td);
	if (!rbt->how->power)
	 { printf ("Turn on arm power\n");
	   while (!rbt->how->power)
	    { nap (0.2);
	    }
	 }

	tbuf->go = (bufp-tbuf->buf)/numJoints;

	while (tbuf->go)
	 { nap (0.2);
	 }

	rciClose();
	exit (0);
}
