#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <stdio.h>

#include <ati.h>

#define MAPPED_IO	0
#define ATI_SPIN_TIMEOUT	10000

/* Only one device per program for now ... */
static AtiDev atid;

static void atiReadData (AtiForceRec *rec, AtiDev *atid);
static int atiReadCom (int com, int mode, void *ptr, int n, AtiDev *atid);
static int atiWriteCom (int com, int mode, void *ptr, int n, AtiDev *atid);

#if MAPPED_IO
#define OUTB(p,v)	regs->p = (v)
#define OUTW(p,v)	regs->p = (v)
#define INB(p)		regs->p
#define INW(p)		regs->p
#else
#define OUTB(p,v)	outb(atip->iobase+ATI_##p,v)
#define OUTW(p,v)	outw(atip->iobase+ATI_##p,v)
#define INB(p)		inb(atip->iobase+ATI_##p)
#define INW(p)		inw(atip->iobase+ATI_##p)
#endif

static void atiReadData (AtiForceRec *rec, AtiDev *atip)
{
#if MAPPED_IO
	AtiRegs *regs = atip->regs;
#endif
	int i;

	OUTW (addr, 0x101);
	rec->dword = INW (data);
	if (rec->dword)
	 { OUTW (addr, 0x101);
	   OUTW (data, 0);	/* clear monitor condition */
	 }
	for (i=0; i<6; i++)
	 { rec->fdata[i] = INW (data);
	 }
}

static int atiReadCom (int com, int mode, void *ptr, int n,
			AtiDev *atip)
{
#if MAPPED_IO
	AtiRegs *regs = atip->regs;
#endif
	unsigned short *dbuf = (unsigned short *)ptr;
	int retcode;

	if (atiWriteCom (com, mode, NULL, 0, atip) < 0)
	 { return (-1);
	 }
	OUTW(addr,0);	  /* reset address */
	retcode = INW(data);
	if (mode != ATI_NO_MODE)
	 { INW(data);
	 }
	while (n--)
	 { *dbuf++ = INW(data);
	 }
	return (retcode);
}

static int atiWriteCom (int com, int mode, void *ptr, int n,
			AtiDev *atip)
{
#if MAPPED_IO
	AtiRegs *regs = atip->regs;
#endif
	unsigned short *dbuf = (unsigned short *)ptr;
	unsigned dword, retcode;
	int spin;

	/*
	   write command block to DPRAM at offset 0x000 
	 */
	OUTW(addr,0); 		/* set address register to command_block */
	OUTW(data,com);
	if (mode != ATI_NO_MODE)
	 { OUTW(data,mode);
	 }
	if (dbuf)
	 { while (n--)
	    { OUTW(data,*dbuf++);
	    }
	 }
	/*
	   read MailBox at DPRAM 0x3FE to clear host interrupt flag
	 */
	OUTW(addr,0x3FE);	     	/* set mailbox address */
	INW(data);		   	/* reset interrupt flag */
	OUTW(addr,0x3FE) ;	    	/* set mailbox address */
	OUTW(data,com);		    	/* clear acknowledgment */

	/*
	   write command word to Mailbox at DPRAM 0x3FF to set DSP intr flag
	 */
	OUTW(addr,0x3ff) ;
	OUTW(data,com) ;

	/*
	   waits for command completion (interrupt flag) from DSP
	 */
	retcode = 1;
	for (spin = ATI_SPIN_TIMEOUT; spin; spin--)
	 {
	   dword = INW(status);
	   // wait for interrupt flag
	   if((dword & 0x0010)==0)     		/* interrupt set */
	    {
	      /* read mailbox location at 0x3FE to reset interrupt flag */
	      OUTW(addr,0x3FE) ;	 	/* write dpram address */
	      retcode = INW(data) ;	 	/* reset interrupt flag */

	      /* test mailbox for command acknowledgment */
	      if(retcode & 0x00ff)		/* test for ack */
	       { break;
	       }
	    }
	 }
	if (spin)
	 { return retcode;
	 }
	else
	 { return -1;
	 }
}

int atiSensorOpen (char *devname, AtiForceRec *fbuf,
		   int n, int m, int srate)
{
	int fd;

	atid.fd = 0;
	if ((fd = open (devname, O_RDWR)) < 0)
	 { return -1;
	 }
	if (ioctl (fd, ATIIOC_GET_SENSOR_INFO, &atid.srec) < 0)
	 { close (fd);
	   return -1;
	 }
	if (ioctl (fd, ATIIOC_SET_CAPBUF_SIZE, m) < 0)
	 { close (fd);
	   return -1;
	 }
printf ("units=%d %d %d \n", atid.srec.units,
			     atid.srec.cntsPerForce, atid.srec.cntsPerTorque);
	atid.fd = fd;
	atid.fbuf = fbuf;
	atid.fsize = n;
	atid.srate = srate;
	return (fd);
}

void atiSensorClose (void)
{
	if (atid.fd)
	 { close (atid.fd);
	 }
	atid.fd = 0;
}

int atiSensorRead (float *f, int *moncode)
{
	/* Reads the current force sensor values into 'f'. If
	   the sensor kernel interface has been opened, these values,
	   plus previously acquired values, are also copied into
	   an internal buffer, where they can be accessed via
	   'atiSensorGetForce'
	 */
	   
	AtiDev *atip = &atid;
	double tmul = 1.0;

	if (atid.srec.units == 3)
	 { tmul = 1000;
	 }
	else if (atid.srec.units != 2)
	 { fprintf (stderr, "atiSensorRead: units=%d\n", atid.srec.units);
	   exit (1);
	 }

	if (atid.fd == 0)
	 { AtiForceRec rec;

	   atid.iobase = 0x280;	/* XXX XXX */
#if 1
	   OUTW (addr, 0x100);
	   if (INW (data))
	    { 
	      atiReadData (&rec, atip);
	      OUTW (addr, 0x100);
	      OUTW (data, 0);

	      f[0] = 10*rec.fdata[0]/(double)atid.srec.cntsPerForce;
	      f[1] = 10*rec.fdata[1]/(double)atid.srec.cntsPerForce;
	      f[2] = 10*rec.fdata[2]/(double)atid.srec.cntsPerForce;

	      f[3] = 10*rec.fdata[3]/(double)atid.srec.cntsPerTorque*tmul;
	      f[4] = 10*rec.fdata[4]/(double)atid.srec.cntsPerTorque*tmul;
	      f[5] = 10*rec.fdata[5]/(double)atid.srec.cntsPerTorque*tmul;
	      if (moncode) *moncode = rec.dword;
	      return 1;
	    }
	   else
	    { return 0;
	    }
#else
	   f[0] = 0;
	   f[1] = 0;
	   f[2] = 0;
	   f[3] = 0;
	   f[4] = 0;
	   f[4] = 0;
	   return 1;
#endif
	 }
	else
	 { AtiForceRec rec;
	   AtiForceRec *fbuf;
	   int nmax;
	   int n;

	   if ((fbuf = atid.fbuf) == NULL)
	    { fbuf = &rec;
	      nmax = 1;
	    }
	   else
	    { nmax = atid.fsize;
	    }
	   rciSaveContext();
	   n = read (atid.fd, fbuf, nmax*sizeof(AtiForceRec));
	   rciRestoreContext();
	   if (n < 0)
	    { return -1;
	    }
	   n /= sizeof(AtiForceRec);
	   atid.num = n;
	   if (n == 0)
	    { return (0);
	    }
	   else
	    { fbuf = &fbuf[n-1];
	      f[0] = 10*fbuf->fdata[0]/(double)atid.srec.cntsPerForce;
	      f[1] = 10*fbuf->fdata[1]/(double)atid.srec.cntsPerForce;
	      f[2] = 10*fbuf->fdata[2]/(double)atid.srec.cntsPerForce;

	      f[3] = 10*fbuf->fdata[3]/(double)atid.srec.cntsPerTorque*tmul;
	      f[4] = 10*fbuf->fdata[4]/(double)atid.srec.cntsPerTorque*tmul;
	      f[5] = 10*fbuf->fdata[5]/(double)atid.srec.cntsPerTorque*tmul;
/* printf ("%d %d %d\n", fbuf->fdata[0], fbuf->fdata[1], fbuf->fdata[2]); */
	      if (moncode) *moncode = fbuf->dword;
	      return 1;
	    }
	 }
}

int atiSensorStart (void)
{
	/* Turns on the force sensor, using appropriate services
	   depending on whether the sensor kernel interface has
	   been opend or not */

	if (atid.fd == 0)
	 { atid.iobase = 0x280;	/* XXX XXX */
	   /* then read the sensor data stuff */
	   if (atiReadCom (ATI_SENSOR_DATA_CMD, ATI_READ, &atid.srec,
			       sizeof(atid.srec)/2, &atid) < 0)
	    { return -1;
	    }
	   rciPrintf ("I\n");
	   return 1;
	 }
	else
	 { int status;
	   rciSaveContext();
	   status = ioctl (atid.fd, ATIIOC_SET_CAPTURE, atid.srate);
	   rciRestoreContext();
	   return (status);
	 }
}

int atiSensorStop (void)
{
	/* Turns off the force sensor, using appropriate services
	   depending on whether the sensor kernel interface has
	   been opened or not */

	if (atid.fd == 0)
	 { atid.iobase = 0x280;	/* XXX XXX */
	   /* then nothing to do */
	   return 1;
	 }
	else
	 { int status;
	   rciSaveContext();
	   status = ioctl (atid.fd, ATIIOC_SET_CAPTURE, 0);
	   rciRestoreContext();
	   return (status);
	 }
}

int atiSensorGetNum ()
{
	/* Returns the number of force sensor value sets that
	   are currently being stored. This number is set when
	   atiSensorRead() is called. */

	return (atid.num);
}

int atiSensorGetForce (float *f, int i)
{
	/* Returns the i-th force sensor value sets, as
	   read in by a previous call to atiSensorRead().
	 */

	AtiForceRec *frec;
	double tmul = 1.0;

	if (atid.srec.units == 3)
	 { tmul = 1000;
	 }

	if (i < atid.num && atid.fbuf)
	 { frec = &atid.fbuf[i];
	   f[0] = 10*frec->fdata[0]/(double)atid.srec.cntsPerForce;
	   f[1] = 10*frec->fdata[1]/(double)atid.srec.cntsPerForce;
	   f[2] = 10*frec->fdata[2]/(double)atid.srec.cntsPerForce;

	   f[3] = 10*frec->fdata[3]/(double)atid.srec.cntsPerTorque*tmul;
	   f[4] = 10*frec->fdata[4]/(double)atid.srec.cntsPerTorque*tmul;
	   f[5] = 10*frec->fdata[5]/(double)atid.srec.cntsPerTorque*tmul;
	   return 1;
	 }
	else
	 { return -1;
	 }
}
