#include <stdio.h>
#include <cdefs.h>
#include <robotMath.h>
#include <robot_kynvar.h>

#define robotName "caesar"

#define EPSILON 1e-6

main()
{
       	/* simple routine to test the kinematics routines 
	   as compiled up for the transputers. */

	TRSF T6;				/* robot T6 transform */
	static float j6[6] = 			/* robot joint angles */
		{ 0.0, -90.0*DEGTORAD, 0.0, 0.0, 45.0*DEGTORAD, 0.0 };
	float j6Check[6];			/* re-computed joint angles */
	JLS *jls;				/* robot joint-level info */
	KYN *kyn;				/* robot kinematic info */
	unsigned long conf;			/* robot solution codes */
	int status, j;
	int errorcnt = 0;

	/* get joint-level information */
	if ((jls = getJls(robotName)) == NULL)
	 { printErrors();
	   exit (1);
	 }
	/* get kinematic information */
	if ((kyn = getKyn(robotName, (int*)0)) == NULL)
	 { printErrors();
	   exit (1);
	 }
	/* compute T6 and conf corresponding to j6 */
	fwdKinematics (&T6, &conf, j6, kyn, (void*)0);
	/* recompute j6 from T6 and conf */
	invKinematics (j6Check, &T6, conf, (float*)0, kyn, jls, (void*)0);

	/* check that we get the same joint angles back */
	for (j=0; j<6; j++)
	 { if (ABS(j6Check[j]-j6[j]) > EPSILON)
	    { fprintf (stderr, "Error, j6Check[%d]=%g\n", j, 
				(double)RADTODEG*j6Check[j]);
	      errorcnt++;
	    }
	 }
	if (errorcnt == 0)
	 { printf ("Looks OK\n");
	 }
}
