/*-=-=-=-=-=-=-=-=-=-=-=- RCCL/RCI Notice of Copyright -=-=-=-=-=-=-=-=-=-=-*
 *									    *
 *	Copyright (C) 1987, 1991 by John E. Lloyd.			    *
 *									    *
 *	Information on copyright permissions, as well as a complete	    *
 *	disclaimer of warranty, is given in the file COPYRIGHT.		    *
 *									    *
 *-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=*/
/*---------------------------------------------------------------------------
 *		
 * MODULE NAME:	simSupport
 *
 * DESCRIPTION:	Some subroutines to make interfacing an RCI task to a robot
 *		simulator program easier.
 *
 * AUTHOR/DATE:	John Lloyd, Fall 1987
 *
 * MODIFICATION HISTORY:
 *
 *--------------------------------------------------------------------------*/

#include <signal.h>
#include "../h/stringDefs.h"
#include <sys/time.h>

#include "../h/cdefs.h"
#include "../h/rciTask.h"
#include "../h/rciMacros.h"
#include "../h/IOdevice.h"

simAttachRobot (td, robotName, simAddr)
RCI_DESC *td;
const char *robotName;
const char *simAddr;
{
	/* used in place of 'rciAttachRobot()': attaches a robot
	   to an RCI task, but specifies a simulator connection as 
	   the I/O device for that robot.
	 */

	char *fxnName = "simAttachRobot()";
	struct devAddr devAddr;
	char simname[256];
	char hostname[256];
	char tagname[256];
	char *user;
	int port;

	if (!planningLevelCheck(fxnName))
	 { return (-1);
	 }
	devAddr.name = "sim";
	devAddr.address = find_robot(robotName);

	if (!simAddr)
	 { 
#if defined(Vxw_TARGET)
	   char ubuf[256];
	   remCurIdGet (ubuf, 0);
	   user = ubuf;
#else
	   user = (char*)getenv("USER");
#endif
	   sprintf (simname, "%s:%s", robotName, user ? user : "0");
	 }
	else
	 { parseSimAddr (tagname, hostname, &port, simAddr);
	   if (port != -1)
	    { sprintf (simname, "%s:%s@%s<%d>", 
		       robotName, tagname, hostname, port);
	    }
	   else
	    { sprintf (simname, "%s:%s@%s", robotName, tagname, hostname);
	    }
	 }
	devAddr.addrStr = simname;
	return (rciAttachRobot (td, robotName, &devAddr));
}

simScheduling (td)
RCI_DESC *td;
{
	/* Used in place of rciScheduling(): selects the appropriate
	   scheduling to run an RCI task off of a simulator.  
	 */

	char *fxnName = "simScheduling()";

	if (!planningLevelCheck(fxnName))
	 { return (-1);
	 }
	if (rciScheduling (td, ON_SIGNAL, SIGALRM) < 0)
	 { return (-1);
	 }
	return (1);
}


