#line 1 "../genConvert.c"
 








#line 1 "../../h/cdefs.h"
 
#line 11 "../genConvert.c"
#line 1 "../../h/robot_jls.h"
 
#line 13
 

#line 1 "../../h/../h/maxjnts.h"
 
#line 16 "../../h/robot_jls.h"



typedef struct _JLS {

  

 int class;  
 int type;  

  

 int numJ;  
 int numFC;  
 int numIC;  
 int forCouplings [3][2];  
 int invCouplings [3][2];  
 float forGearRatios [12+3];  
 float invGearRatios [12+3];  

  

 float angtoenc [12+3];  
 float enctoang [12+3];  
 int encOffset [12];  

  

 float mtorqueToCurrent [12];  
 float currentToMtorque [12];  
 int currentBias [12];  
  

  

 int encWheelSize [12];  
 int encCalvec [12];  
 int encHome [12];  
 float potSlope [12];  
 float potIntercept [12];  

  

 char jtype [12+1];  
 float jmin [12];  
 float jmax [12];  
 float jrng [12];  
 float maxVel [12];  
 float maxAcc [12];  
 float maxJerk[12];  

 float maxMtorque [12];  
 float maxJtorque [12];  

  

 float parkAngles [12];  
 float bestAngles [12];  
 float averageGearRatio;  
 float averageAngtoenc;  
 float nominalVel [12];  
 float nominalAcc [12];  
 float nominalJerk[12];  
 float averageVel;  
 float averageAcc;  
 float averageJerk;  
 long flags;  

  

 int (*encoderToAngle)();  
 int (*angleToEncoder)();  
 int (*encoderToJoint)();  
 int (*jointToEncoder)();  
 int (*mtorqueToJtorque)();  
 int (*jtorqueToMtorque)();  
} JLS;

typedef struct {
 struct _GEN_KYN *gen;
 void *ext;
} KYN;









extern JLS *getJls (const char*);
extern KYN *getKyn (const char*, int*);
extern find_robot (const char*);
extern robot_class (const char*);
extern int getVarSize (const char*);
extern char *getDefaultRobot ();
extern getRobotPosition (float*, const char*, const char*);
#line 12 "../genConvert.c"
#line 1 "../../h/robotMath.h"
 
#line 13 "../genConvert.c"

encoderToAngle (ang, enc, jls)
register float ang[];
register int enc[];
register JLS *jls;
{
 int e[12];
 int nj = jls->numJ;
 int nc = jls->numIC;

 register int j, k;
 register int *c = (int *)jls->invCouplings;

 for (j=0; j<nj; j++)
 { ang[j] = jls->enctoang[j] * (e[j] = enc[j] - jls->encOffset[j]);
 }
 for (k=0; k<nc; k++)
 { ang[c[0]] += jls->enctoang[j++] * e[c[1]];
 c += 2;
 }
}

angleToEncoder (enc, ang, jls)
register int enc[];
register float ang[];
register JLS *jls;
{
 float val;
 int nj = jls->numJ;
 int nc = jls->numFC;

 register int j, k;
 register int *c = (int *)jls->forCouplings;

 for (j=0; j<nj; j++)
 { val = jls->angtoenc[j] * ang[j] + jls->encOffset[j];
 enc[j] = ((int)(((val)<0) ? ((val)-0.5) : ((val)+0.5)));
 }
 for (k=0; k<nc; k++)
 { enc[c[0]] += jls->angtoenc[j++] * ang[c[1]];
 c += 2;
 }
}

encoderToJoint (ang, enc, jls)
register float ang[];
register int enc[];
register JLS *jls;
{
 int nj = jls->numJ;
 int nc = jls->numIC;

 register int j, k;
 register int *c = (int *)jls->invCouplings;

 for (j=0; j<nj; j++)
 { ang[j] = jls->enctoang[j] * enc[j];
 }
 for (k=0; k<nc; k++)
 { ang[c[0]] += jls->enctoang[j++] * enc[c[1]];
 c += 2;
 }
}

jointToEncoder (enc, ang, jls)
register int enc[];
register float ang[];
register JLS *jls;
{
 float val;
 int nj = jls->numJ;
 int nc = jls->numFC;

 register int j, k;
 register int *c = (int *)jls->forCouplings;

 for (j=0; j<nj; j++)
 { val = jls->angtoenc[j] * ang[j];
 enc[j] = ((int)(((val)<0) ? ((val)-0.5) : ((val)+0.5)));
 }
 for (k=0; k<nc; k++)
 { enc[c[0]] += jls->angtoenc[j++] * ang[c[1]];
 c += 2;
 }
}

mtorqueToJtorque (jt, mt, jls)
register float jt[];
register float mt[];
JLS *jls;
{
 int nj = jls->numJ;
 int nc = jls->numFC;

 register int j, k;
 register int *c = (int *)jls->forCouplings;

 for (j=0; j<nj; j++)
 { jt[j] = jls->forGearRatios[j] * mt[j];
 }
 for (k=0; k<nc; k++)
 { jt[c[1]] += jls->forGearRatios[j++] * mt[c[0]];
 c += 2;
 }
}

jtorqueToMtorque (mt, jt, jls)
register float mt[];
register float jt[];
JLS *jls;
{
 int nj = jls->numJ;
 int nc = jls->numIC;

 register int j, k;
 register int *c = (int *)jls->invCouplings;

 for (j=0; j<nj; j++)
 { jt[j] = jls->invGearRatios[j] * mt[j];
 }
 for (k=0; k<nc; k++)
 { jt[c[1]] += jls->invGearRatios[j++] * mt[c[0]];
 c += 2;
 }
}

radianToDegree (deg, rad, jls)
float *deg, *rad;
JLS *jls;
{
 register int j;

 for (j=0; j<jls->numJ; j++)
 { if (jls->jtype[j] == 'R')
 { deg[j] = 57.29577951308232100*rad[j];
 }
 else
 { deg[j] = rad[j];
 }
 }
}

degreeToRadian (rad, deg, jls)
float *rad, *deg;
JLS *jls;
{
 register int j;

 for (j=0; j<jls->numJ; j++)
 { if (jls->jtype[j] == 'R')
 { rad[j] = 0.01745329251994330*deg[j];
 }
 else
 { rad[j] = deg[j];
 }
 }
}
