/*-=-=-=-=-=-=-=-=-=-=-=- RCCL/RCI Notice of Copyright -=-=-=-=-=-=-=-=-=-=-*
 *									    *
 *	Copyright (C) 1987, 1991 by John E. Lloyd.			    *
 *									    *
 *	Information on copyright permissions, as well as a complete	    *
 *	disclaimer of warranty, is given in the file COPYRIGHT.		    *
 *									    *
 *-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=*/

#include <stdio.h>
#include "../h/stringDefs.h"

#if 0
#include <sys/types.h>
#include <sys/stat.h>
#endif

#define COMPILED_JLS	0

#include "../h/cdefs.h"
#include "../h/robot_jls.h"
#include "../h/robot_kynvar.h"
#include "../h/robotComfxns.h"
#include "../h/robotMath.h"
#include "../h/errorCodes.h"
#include "../h/errorStack.h"
#include "../h/utilFxns.h"

#include "robotData.h"

static ROBOT_DATA_PTRS *robotDataPtrs;
static int tableSize;

void *malloc();
extern char *rciGetenv();

/* ----- Error Messages. Defined externally to save space ----- */

static char *noMemoryMsg	= "%s -- out of memory\n";
static char *noEnvMsg		= "%s -- environment variable %s not set\n";
static char *cantCreateMsg	= "%s -- can't create structure for '%s'\n";
static char *noCreateFxnMsg	= "%s -- no create function for '%s'\n";
static char *rbtNotDefMsg	= "%s -- rbt %s: not defined in RCI tables\n";
static char *noStatBuildMsg	= "%s -- can't stat or build '%s'\n";
static char *badFileMsg	  	= "%s -- error reading file '%s'\n";
static char *badChecksumMsg  	= "%s -- checksum error in file '%s'\n";
static char *fileTooLongMsg  	= "%s -- file '%s' is too long\n";
static char *cantAccessFileMsg	= "%s -- can't access file %s\n";

#if COMPILED_JLS
static JLS *getCompiledJls ();
#else
static JLS *getAsciiJls ();
#endif

static initDone = 0;
static robotDataSetup()
{
	ROBOT_DATA_ENTRY *rde;
	ROBOT_DATA_PTRS *ptr;
	int i;

	if (robotDataPtrs)
	 { FREE(robotDataPtrs);
	 }
	tableSize = 0;
	for (rde=rciRobotDataTable; rde->name; rde++)
	 { tableSize++;
	 }
	if (tableSize > 0)
	 { ptr = (ROBOT_DATA_PTRS*)malloc(sizeof(ROBOT_DATA_PTRS)*tableSize);
	 }
	for (i=0; i<tableSize; i++)
	 { ptr[i].jls = NULL;
	   ptr[i].kyn = NULL;
	 }
	robotDataPtrs = ptr;
	initDone = 1;
	requestFlagInit(&initDone);
}

#if 0

find_robot (name)
const char *name;
{
	register ROBOT_DATA_ENTRY *rde;
	register int i;

	for (rde=rciRobotDataTable,i=1; rde->name; rde++,i++)
	 { if (strcmp (name, rde->name) == 0)
	    { return (i);
	    }
	 }
 	return (0);
}

robot_class(name)
const char *name;
{
	register ROBOT_DATA_ENTRY *rde;

	for (rde=rciRobotDataTable; rde->name; rde++)
	 { if (strcmp (name, rde->name) == 0)
	    { return (rde->rbtclass);
	    }
	 }
 	return (0);
}

get_comfxns (fxns, name)
ROBOT_COMFXNS *fxns;
char *name;
{
	register ROBOT_DATA_ENTRY *rde;

	for (rde=rciRobotDataTable; rde->name; rde++)
	 { if (strcmp (name, rde->name) == 0)
	    { fxns->startup = rde->startupFxn;
	      fxns->input =   rde->inputFxn;
	      fxns->output =  rde->outputFxn;
	      fxns->release = rde->releaseFxn;
	      return (1);
	    }
	 }
	return (0);
}

#endif

KYN *doGetKyn (name, size, fxn)
const char *name;
int size;
int (*fxn)();
{
	char *fxnName = "doGetKyn()";

	char *fileName = NULL;
	char baseName[128];
	FILE *fp = NULL;
	char *rootdir;
        KYN *kyn = NULL;
	int j;

	sprintf (baseName, "robotData/%s.kyn", name);
	if ((fp = locateFile (&fileName, baseName)) == NULL)
	 { errorMsg (EBadAccess, cantAccessFileMsg, fxnName, baseName);
	   goto error;
	 }
	if ((kyn = ALLOC(KYN)) == NULL)
	 { errorMsg (EMemory, noMemoryMsg, fxnName);
	   goto error;
	 }
	kyn->ext = NULL;
	kyn->gen = NULL;
	if ((kyn->ext = (void*)malloc(size)) == NULL ||
	    (kyn->gen = ALLOC(GEN_KYN)) == NULL)
	 { errorMsg (EMemory, noMemoryMsg, fxnName);
	   goto error;
	 }
	BZERO(kyn->gen,sizeof(GEN_KYN));
	/* paranoid ... bzero doesn't guarantee to set floats to 0.0 */
	for (j=0; j<MAXJNTS; j++)
	 { kyn->gen->dh[j].theta = 0.0;
	   kyn->gen->dh[j].d = 0.0;
	   kyn->gen->dh[j].a = 0.0;
	   kyn->gen->dh[j].alpha = 0.0;
	   kyn->gen->dh[j].beta = 0.0;
	 }

	/* call class specific routine to do the work */

	if ((*fxn)(kyn,fp) == -1)
	 { errorMsg (getErrorCode(0), badFileMsg, fxnName, fileName);
	   goto error;
	 }
	free (fileName);
  	fclose (fp);
 	return (kyn);
error:
	if (fileName)
	 { free (fileName);
	 }
	if (fp)
	 { fclose (fp);
	 }
	if (kyn)
	 { if (kyn->gen)
	    { FREE (kyn->gen);
	    }
	   if (kyn->ext)
	    { FREE (kyn->ext);
	    }
	   FREE (kyn);
	 }
	return (NULL);
}

KYN *getKyn (name, sizeptr)
const char *name;
int *sizeptr;
{
	char *fxnName = "getKyn()";
	register ROBOT_DATA_ENTRY *rde;
	register ROBOT_DATA_PTRS *ptr;

	if (!planningLevelCheck(fxnName))
	 { return (NULL);
	 }
	if (!initDone)
	 { robotDataSetup();
	 }

	ptr=robotDataPtrs;
	for (rde=rciRobotDataTable; rde->name; rde++,ptr++)
	 { if (strcmp (name, rde->name) == 0)
	    { if (ptr->kyn)
	       { if (sizeptr)
		  { *sizeptr = rde->kynSize;
		  }
		 return (ptr->kyn);
	       }
	      else
	       { if (rde->getKyn)
		  { KYN *kyn;
		    if ((kyn=doGetKyn(name,rde->kynSize,rde->getKyn)) == NULL)
		     { errorMsg (getErrorCode(0), cantCreateMsg, fxnName, name);
		       return (NULL);
		     }
		    if (sizeptr)
		     { *sizeptr = rde->kynSize;
		     }
		    ptr->kyn = kyn;
		    return (kyn);
		  }
		 else
	          { errorMsg (EBadFxn, noCreateFxnMsg, fxnName, name);
		    return (NULL);
	          }
	       }
	    }
	 }
	errorMsg (ENotDefined, rbtNotDefMsg, fxnName, name);
	return (NULL);
}

#if 0

int getVarSize (name)
const char *name;
{
	char *fxnName = "getVarSize()";
	register ROBOT_DATA_ENTRY *rde;

	if (!planningLevelCheck(fxnName))
	 { return (-1);
	 }
	if (!initDone)
	 { robotDataSetup();
	 }
	for (rde=rciRobotDataTable; rde->name; rde++)
	 { if (strcmp (name, rde->name) == 0)
	    { return (rde->varSize);
	    }
	 }
	errorMsg (ENotDefined, rbtNotDefMsg, fxnName, name);
	return (-1);
}

#endif

#if COMPILED_JLS

static JLS *getCompiledJls (name)
const char *name;
{
	char *fxnName = "getCompiledJls()";

#define FNS 256
	char fileName_JLS[FNS+1];
	char fileName_jls[FNS+1];
	FILE *fp = NULL;
	char *rootdir;
        JLS *jls = NULL;
	int checksum;
	struct stat stat_jls, stat_JLS;
	int statValid_jls, statValid_JLS;

	if ((jls = (JLS*)malloc(sizeof(JLS))) == NULL)
	 { errorMsg (EMemory, noMemoryMsg, fxnName);
	   goto errorReturn;
	 }
	if ((rootdir = (char *)rciGetenv("RCCL_CONF")) == NULL)
  	 { errorMsg (EEnviron, noEnvMsg, fxnName, "RCCL_CONF");
	   goto errorReturn;
	 }
	sprintf (fileName_jls, "%s/robots/%s.jls", rootdir, name);
	sprintf (fileName_JLS, "%s/robots/%s.JLS", rootdir, name);

	if (stat (fileName_JLS, &stat_JLS) < 0)
	 { statValid_JLS = 0;
	 }
	else
	 { statValid_JLS = 1;
	 }
	if (stat (fileName_jls, &stat_jls) < 0)
	 { statValid_jls = 0;
	 }
	else
	 { statValid_jls = 1;
	 }

	if (statValid_jls)
	 { if ((statValid_JLS && stat_jls.st_mtime > stat_JLS.st_mtime) ||
	       !statValid_JLS)
	    { char comstring[1024];
#ifdef sun
	      /* XXX do this because sunTouchMemoryHack() will bring
		 in rciSimOnly. */
	      extern int rciSimOnly;
	      int forceRef;

	      forceRef = rciSimOnly;
#endif
	      printf ("Rebuilding %s ...\n", fileName_JLS);
	      sprintf (comstring, "makejls %s %s", fileName_jls,
						     fileName_JLS);
	      printf ("%s\n", comstring);
	      if (system (comstring) != 0)
	       { sysErrorMsg (ENoFile, fxnName, "system");
		 goto errorReturn;
	       }
#ifdef sun
	      sunTouchMemoryHack();	/* XXXXXX !!!!!! */
#endif
	    }
	 }
	else if (!statValid_JLS)
	 { errorMsg (ENoFile, noStatBuildMsg, fxnName, fileName_JLS);
	   goto errorReturn;
	 }

	if ((fp = fopen (fileName_JLS, "r")) == NULL)
	 { sysErrorMsg (ENoFile, fxnName, "fopen on JLS file");
	   goto errorReturn;
	 }
	if ((fread (jls, sizeof(JLS), 1, fp) != 1) ||
	    (fread (&checksum, sizeof (long), 1, fp) != 1) )
	 { errorMsg (EBadFile, badFileMsg, fxnName, fileName_JLS);
	   goto errorReturn;
	 }

	/* ---- Compute checksum ---- */

	 { register int i, x = 0, *p = (int *) jls;

	   for (i=0; i<sizeof(JLS)/sizeof(int); i++)
	    { x += p[i];
	    }
	   if (x != checksum)
	    { errorMsg (EBadFile, badChecksumMsg, fileName_JLS);
	      goto errorReturn;
	    }
	 }

	/* ---- Now try to read just one more byte ---- */

	 { char c;
	   if (fread (jls, sizeof(char), 1, fp) == 1)
	    { errorMsg (EBadFile, fileTooLongMsg, fxnName, fileName_JLS);
	      goto errorReturn;
	    }
	 }

	fclose (fp);
 	return (jls);
errorReturn:
	if (fp)
	 { fclose (fp);
	 }
	if (jls)
	 { FREE(jls);
	 }
	return (NULL);
}

#else

static JLS *getAsciiJls (name)
const char *name;
{
	char *fxnName = "getAsciiJls()";

	char *fileName = NULL;
	char baseName[128];
	char *rootdir;
	FILE *fp = NULL;
        JLS *jls = NULL;

	if ((jls = (JLS*)malloc(sizeof(JLS))) == NULL)
	 { errorMsg (EMemory, noMemoryMsg, fxnName);
	   goto errorReturn;
	 }
	sprintf (baseName, "robotData/%s.jls", name);
	if ((fp = locateFile (&fileName, baseName)) == NULL)
	 { errorMsg (EBadAccess, cantAccessFileMsg, fxnName, baseName);
	   goto errorReturn;
	 }
	if (readJls (jls, fp) < 0)
	 { errorMsg (EBadFile, badFileMsg, fxnName, fileName);
	   goto errorReturn;
	 }
	free (fileName);
	fclose (fp);
 	return (jls);
errorReturn:
	if (fileName)
	 { free (fileName);
	 }
	if (fp)
	 { fclose (fp);
	 }
	if (jls)
	 { FREE(jls);
	 }
	return (NULL);
}

#endif

JLS *getJls (name)
const char *name;
{
#if COMPILED_JLS
#define GET_JLS_FXN	getCompiledJls
#else
#define GET_JLS_FXN	getAsciiJls
#endif

	char *fxnName = "getJls()";
	register ROBOT_DATA_ENTRY *rde;
	register ROBOT_DATA_PTRS *ptr;

	if (!planningLevelCheck(fxnName))
	 { return (NULL);
	 }
	if (!initDone)
	 { robotDataSetup();
	 }
	ptr=robotDataPtrs;
	for (rde=rciRobotDataTable; rde->name; rde++,ptr++)
	 { if (strcmp (name, rde->name) == 0)
	    { if (ptr->jls)
	       { return (ptr->jls);
	       }
	      else
	       { if (rde->getJls)
		  { JLS *jls;
		    if ((jls = GET_JLS_FXN(name)) == NULL)
		     { errorMsg (EBadFile, cantCreateMsg, fxnName, name);
		       return (NULL);
		     }
		    /* call class function to instantiate function pointers */
		    (*rde->getJls)(jls);

		    ptr->jls = jls;
		    return (jls);
		  }
		 else
	          { errorMsg (EBadFxn, noCreateFxnMsg, fxnName, name);
		    return (NULL);
	          }
	       }
	    }
	 }
	errorMsg (ENotDefined, rbtNotDefMsg, fxnName, name);
	return (NULL);
}

int gen_readNomVelAcc (char *fxnName, GEN_KYN *gen)
{
	char buf[256];
	double *resp;

	gen->nominalTransVel = -1;
	gen->nominalRotVel = -1;
	gen->nominalTransAcc = -1;
	gen->nominalRotAcc = -1;

	while (readString (buf, 256))
	 { 
	   if (strcmp (buf, "NOM_TRANS_VEL") == 0)
	    { resp = &gen->nominalTransVel;
	    }
	   else if (strcmp (buf, "NOM_ROT_VEL") == 0)
	    { resp = &gen->nominalRotVel;
	    }
	   else if (strcmp (buf, "NOM_TRANS_ACC") == 0)
	    { resp = &gen->nominalTransAcc;
	    }
	   else if (strcmp (buf, "NOM_ROT_ACC") == 0)
	    { resp = &gen->nominalRotAcc;
	    }
	   else
	    { errorMsg (EBadFile,
"%s -- bogus field %s in file\n", fxnName, buf);
	      return (-1);
	    }
	   if (!readExactString ("="))
	    { errorMsg (EBadFile, 
"%s -- field %s should be followed by '='\n", fxnName, buf);
	      return (-1);
	    }
	   if (!readDouble (resp))
	    { errorMsg (EBadFile, 
"%s -- can't read number following %s\n", fxnName, buf);
	      return (-1);
	    }
	 }
	if (gen->nominalRotVel != -1)
	 { gen->nominalRotVel *= DEGTORAD;
	 }
	if (gen->nominalRotAcc != -1)
	 { gen->nominalRotAcc *= DEGTORAD;
	 }
	return 1;
}

#ifdef Tsp_TARGET		/* XXX */
/* XXX hack to refer to atof early so we get the RCCL/RCI version */
extern double atof();
static double (*ptr)() = atof;
#endif
