/*-=-=-=-=-=-=-=-=-=-=-=- RCCL/RCI Notice of Copyright -=-=-=-=-=-=-=-=-=-=-*
 *									    *
 *	Copyright (C) 1991 by John E. Lloyd.				    *
 *									    *
 *	Information on copyright permissions, as well as a complete	    *
 *	disclaimer of warranty, is given in the file COPYRIGHT.		    *
 *									    *
 *-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=*/
#include <stdio.h>
#include <math.h>

#include "../h/cdefs.h"
#include "../h/rccl.h"	/* XXX MUST include rccl.h before grf4.h */
#include "../h/gantry_kynvar.h"

#include "../h/grf4.h"

#define NUM_CIRC_EDGES		12

#define J1	0
#define J2	1
#define J3	2
#define J4	3
#define J5	4

static int capacity;

#define BEIGE		RGBPACK(225,160, 90)
#define DARK_BROWN	RGBPACK( 85, 55, 35)
#define DARK_GREY	RGBPACK( 50, 50, 50)
#define BLACK		RGBPACK(  0,  0,  0)
#define WHITE		RGBPACK(255,255,255)
#define SILVER		RGBPACK(200,200,220)
#define AMBER		RGBPACK(255,125,  0)

/* Gantry sim stuff */

#define GSIM_BEAM_WIDE			 8.0
#define GSIM_BEAM_THIN			 4.0
#define GSIM_POST_HEIGHT		156.5
#define GSIM_BEAM_X_LEN			204
#define GSIM_BEAM_Y_LEN			182
#define GSIM_LINK_X_LEN			198
#define GSIM_LINK_Z_LEN			84.7
#define GSIM_LINK_XZ_SEP		10.0
#define GSIM_BEAM_LINK_X_SEP		14.0
#define GSIM_PT_RADIUS			10.5
#define GSIM_PT_DEPTH			10.7
#define GSIM_PT_SEP			.8


static multA1();
static multA2();
static multA3();
static multA4();
static multA5();

drawGantryHand ()
{
	grf4Move (0.0, 0.0, 0.0);
	grf4Draw (0.0, 0.0, 2*GSIM_PT_RADIUS);
}

Grf4Object *gantrySidePost;
Grf4Object *gantryBeamX;
Grf4Object *gantryBeamY;
Grf4Object *gantryLinkX;
Grf4Object *gantryLinkZ;
Grf4Object *gantryPanTilt;

defineGantrySimLinks()
{
	void *baseMat;
	void *bodyMat;
	void *wristMat;

	bodyMat = grf4DefMaterial(SILVER, SILVER, 0, 0, 0.0, 1.0);
	wristMat = grf4DefMaterial(DARK_GREY, DARK_GREY, 0,0x808080,50.0,1.0);

	gantrySidePost = grf4GenBox (GSIM_BEAM_WIDE, GSIM_BEAM_WIDE, 
				     GSIM_POST_HEIGHT, bodyMat);
	gantryBeamX = grf4GenBox (GSIM_BEAM_WIDE, GSIM_BEAM_WIDE,
				      GSIM_BEAM_X_LEN, bodyMat);
	gantryBeamY = grf4GenBox (GSIM_BEAM_WIDE, GSIM_BEAM_WIDE,
				      GSIM_BEAM_Y_LEN, bodyMat);
	gantryLinkX = grf4GenBox (GSIM_BEAM_WIDE, GSIM_BEAM_THIN,
				  GSIM_LINK_X_LEN, bodyMat);
	gantryLinkZ = grf4GenBox (GSIM_BEAM_WIDE, GSIM_BEAM_THIN,
				  GSIM_LINK_Z_LEN, bodyMat);
	gantryPanTilt = grf4GenDisc (12, GSIM_PT_RADIUS, GSIM_PT_DEPTH,
					 wristMat);
  	capacity = grf4Capacity();
}

drawGantrySim (T, kyn, var, drawHand, handpos)
register TRSF *T;
KYN *kyn;
void *var;
int drawHand;
int handpos;
{
	/* clear the display and draw a robot, given a set of joint angles */

	GANTRY_KYN *evar = (GANTRY_KYN*)var;
	GEN_KYN *gen = kyn->gen;

	TRSF trans;

	/* Draw the 4 posts */
	grf4GetTrans (&trans);
	T[0] = trans;
	multTrsfXyz (&trans, GSIM_BEAM_X_LEN/2, GSIM_BEAM_Y_LEN/2, 0);
	grf4SetTrans (&trans);
	grf4DrawObject (gantrySidePost);
	multTrsfXyz (&trans, 0, -GSIM_BEAM_Y_LEN-GSIM_BEAM_WIDE, 0);
	grf4SetTrans (&trans);
	grf4DrawObject (gantrySidePost);
	multTrsfXyz (&trans, -GSIM_BEAM_X_LEN-GSIM_BEAM_WIDE, 0, 0);
	grf4SetTrans (&trans);
	grf4DrawObject (gantrySidePost);
	multTrsfXyz (&trans, 0, +GSIM_BEAM_Y_LEN+GSIM_BEAM_WIDE, 0);
	grf4SetTrans (&trans);
	grf4DrawObject (gantrySidePost);

	trans = T[0];
	multTrsfRpy (&trans, 0, -90, 0);
	multTrsfXyz (&trans, -GSIM_BEAM_WIDE/2, -GSIM_BEAM_WIDE/2,
			     -GSIM_BEAM_X_LEN/2);
	multTrsfXyz (&trans, GSIM_POST_HEIGHT-GSIM_BEAM_WIDE/2,
			     (GSIM_BEAM_Y_LEN+GSIM_BEAM_WIDE)/2, 0);
	grf4SetTrans (&trans);
	grf4DrawObject (gantryBeamX);
	multTrsfXyz (&trans, 0, -GSIM_BEAM_Y_LEN-GSIM_BEAM_WIDE, 0);
	grf4SetTrans (&trans);
	grf4DrawObject (gantryBeamX);

	trans = T[0];
	multTrsfRpy (&trans, 0, 0, -90);
	multTrsfXyz (&trans, -GSIM_BEAM_WIDE/2, -GSIM_BEAM_WIDE/2,
			     -GSIM_BEAM_Y_LEN/2);
	multTrsfXyz (&trans, (GSIM_BEAM_X_LEN+GSIM_BEAM_WIDE)/2,
			     -GSIM_POST_HEIGHT+GSIM_BEAM_WIDE/2, 0);
	grf4SetTrans (&trans);
	grf4DrawObject (gantryBeamY);
	multTrsfXyz (&trans, -GSIM_BEAM_X_LEN-GSIM_BEAM_WIDE, 0, 0);
	grf4SetTrans (&trans);
	grf4DrawObject (gantryBeamY);

	multA1 (&trans, &T[0], kyn, evar);
	T[1] = trans;
	multTrsfXyz (&trans, -GSIM_BEAM_WIDE/2, -GSIM_BEAM_THIN/2, 
			     -GSIM_LINK_X_LEN/2);
	multTrsfXyz (&trans, -GSIM_POST_HEIGHT-GSIM_BEAM_LINK_X_SEP-
			      GSIM_BEAM_WIDE/2,
			      GSIM_BEAM_WIDE/2+GSIM_BEAM_THIN/2+
			      GSIM_LINK_XZ_SEP, 0);
	grf4SetTrans (&trans);
	grf4DrawObject (gantryLinkX);
	
	multA2 (&trans, &T[1], kyn, evar);
	T[2] = trans;
	multA3 (&trans, &T[2], kyn, evar);
	T[3] = trans;
	multTrsfXyz (&trans, 0, 0, 
			     -GSIM_PT_RADIUS-GSIM_PT_SEP-GSIM_PT_DEPTH);
	grf4SetTrans (&trans);
	grf4DrawObject (gantryPanTilt);
	multTrsfXyz (&trans, -GSIM_BEAM_WIDE/2, -GSIM_BEAM_THIN/2, 
			     -GSIM_LINK_Z_LEN);
	grf4SetTrans (&trans);
	grf4DrawObject (gantryLinkZ);
	
	multA4 (&trans, &T[3], kyn, evar);
	T[4] = trans;
	multTrsfXyz (&trans, 0, 0, -GSIM_PT_DEPTH/2);
	grf4SetTrans (&trans);
	grf4DrawObject (gantryPanTilt);

	multA5 (&trans, &T[4], kyn, evar);
	T[5] = trans;

	if (drawHand)
	 { grf4SetTrans (&trans);
	   drawGantryHand ();
	 }
}

gantrySimViewSize (r, b, h, kyn)
float *r, *b, *h;
KYN *kyn;
{
	GEN_KYN *gen = kyn->gen;

	*r = GSIM_BEAM_X_LEN;
	*b = GSIM_POST_HEIGHT/2;
	*h = *r;
}

gantrySimLinkFrames (T, B, kyn, var)
TRSF *T, *B;
KYN *kyn;
void *var;
{
	GANTRY_KYN *evar = (GANTRY_KYN*)var;

	T[0] = *B;
	multA1 (&T[1], &T[0], kyn, evar);
	multA2 (&T[2], &T[1], kyn, evar);
	multA3 (&T[3], &T[2], kyn, evar);
	multA4 (&T[4], &T[3], kyn, evar);
	multA5 (&T[5], &T[4], kyn, evar);
}

/* Routines for calculating the frame matrices for GANTRY robots */

static multA1(tr, t1, kyn, var)
register TRSF *tr;
register TRSF *t1;
KYN *kyn;
GANTRY_VAR *var;
{
	register GEN_KYN *gen = kyn->gen;
	float D;

	D = var->d[J1]/10;

	tr->n.x = -t1->a.x;
	tr->o.x = -t1->n.x;
	tr->a.x =  t1->o.x;
	tr->p.x =  t1->n.x*D + t1->p.x;

	tr->n.y = -t1->a.y;
	tr->o.y = -t1->n.y;
	tr->a.y =  t1->o.y;
	tr->p.y =  t1->n.y*D + t1->p.y;

	tr->n.z = -t1->a.z;
	tr->o.z = -t1->n.z;
	tr->a.z =  t1->o.z;
	tr->p.z =  t1->n.z*D + t1->p.z;
}

static multA2(tr, t1, kyn, var)
register TRSF *tr;
register TRSF *t1;
KYN *kyn;
GANTRY_VAR *var;
{
	register GEN_KYN *gen = kyn->gen;
	float D;

	D = var->d[J2]/10;

	tr->n.x = -t1->o.x;
	tr->o.x =  t1->a.x;
	tr->a.x = -t1->n.x;
	tr->p.x =  t1->a.x*D + t1->p.x;

	tr->n.y = -t1->o.y;
	tr->o.y =  t1->a.y;
	tr->a.y = -t1->n.y;
	tr->p.y =  t1->a.y*D + t1->p.y;

	tr->n.z = -t1->o.z;
	tr->o.z =  t1->a.z;
	tr->a.z = -t1->n.z;
	tr->p.z =  t1->a.z*D + t1->p.z;
}

static multA3(tr, t1, kyn, var)
register TRSF *tr;
register TRSF *t1;
KYN *kyn;
GANTRY_VAR *var;
{
	register GEN_KYN *gen = kyn->gen;
	float D;

	D = var->d[J3]/10;

	tr->n.x = -t1->n.x;
	tr->o.x =  t1->o.x;
	tr->a.x = -t1->a.x;
	tr->p.x =  t1->a.x*D + t1->p.x;

	tr->n.y = -t1->n.y;
	tr->o.y =  t1->o.y;
	tr->a.y = -t1->a.y;
	tr->p.y =  t1->a.y*D + t1->p.y;

	tr->n.z = -t1->n.z;
	tr->o.z =  t1->o.z;
	tr->a.z = -t1->a.z;
	tr->p.z =  t1->a.z*D + t1->p.z;
}

static multA4(tr, t1, kyn, var)
register TRSF *tr;
register TRSF *t1;
KYN *kyn;
GANTRY_VAR *var;
{
	register GEN_KYN *gen = kyn->gen;
	register float C;
	register float S;

	C = var->c4;
	S = var->s4;

	tr->n.x = C*t1->n.x + S*t1->o.x;
	tr->o.x = -t1->a.x;
	tr->a.x = -S*t1->n.x + C*t1->o.x;
	tr->p.x = t1->p.x;

	tr->n.y = C*t1->n.y + S*t1->o.y;
	tr->o.y = -t1->a.y;
	tr->a.y = -S*t1->n.y + C*t1->o.y;
	tr->p.y = t1->p.y;

	tr->n.z = C*t1->n.z + S*t1->o.z;
	tr->o.z = -t1->a.z;
	tr->a.z = -S*t1->n.z + C*t1->o.z;
	tr->p.z = t1->p.z;
}

static multA5(tr, t1, kyn, var)
register TRSF *tr;
register TRSF *t1;
KYN *kyn;
GANTRY_VAR *var;
{
	register GEN_KYN *gen = kyn->gen;
	register float C, S;

	C = var->c5;
	S = var->s5;

	tr->n.x =   C*t1->n.x + S*t1->o.x;
	tr->o.x =   t1->a.x;
	tr->a.x =   S*t1->n.x - C*t1->o.x;
	tr->p.x =   t1->p.x;

	tr->n.y =   C*t1->n.y + S*t1->o.y;
	tr->o.y =   t1->a.y;
	tr->a.y =   S*t1->n.y - C*t1->o.y;
	tr->p.y =   t1->p.y;

	tr->n.z =   C*t1->n.z + S*t1->o.z;
	tr->o.z =   t1->a.z;
	tr->a.z =   S*t1->n.z - C*t1->o.z;
	tr->p.z =   t1->p.z;
}
