/*-=-=-=-=-=-=-=-=-=-=-=- RCCL/RCI Notice of Copyright -=-=-=-=-=-=-=-=-=-=-*
 *									    *
 *	Copyright (C) 1987, 1991 by John E. Lloyd and Mike Parker.	    *
 *									    *
 *	Information on copyright permissions, as well as a complete	    *
 *	disclaimer of warranty, is given in the file COPYRIGHT.		    *
 *									    *
 *-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=*/
/*---------McGill-Computer-Vision-and-Robotics-Laboratory--C-Module---------
 *		
 * NAME:	cputime, int_cputime
 *
 *
 * USAGE:
 *		time = cputime (option);
 *		  double time;
 *		  enum cputime_option;
 *
 * DESCRIPTION: Returns the time consumed by a process. The options are as
 *		follows:
 *
 *		  System  : Time spent executing system calls for the process.
 *		  User    : Time spent executing the processes own code.
 *		  All	  : All time spent executing for the process.
 * 
 * NOTES:
 *
 * AUTHOR/DATE: John Lloyd and Mike Parker, September 1984
 *
 * UPDATES:
 *
 *-------------McGill-Computer-Vision-and-Robotics-Laboratory--------------*/

#include "useful.h"

#ifdef Tsp_TARGET
#include <conc.h>

double cputime(option)
enum cputime_option option;
{
	double usecPerTick = 1;

	if (ProcGetPriority())	/* then process is low priority */
	 { usecPerTick = 64;
	 }
	return ((Time()*usecPerTick)/1e6);
}

cputimeInit()
{
}

cputimeExit()
{
}
#else 
#ifdef Dsp_TARGET
#if 0
/* #define CLOCK_PER_SEC	40000000.0 */
#include <timer40.h>

double cputime(option)
enum cputime_option option;
{
	return (elapse());
}

cputimeInit()
{
	time_run();
}

cputimeExit()
{
	(void)time_end();
}
#else
#include <timer40.h>
#include <intpt40.h>

double cputime(option)
enum cputime_option option;
{
	return (time_count / 1e3);
}

cputimeInit()
{
    /*----------------------------------------------------------------------*/
    /*  SET UP TIMER 0 INTERRUPT FOR 64 BIT TIMER COUNTER                   */
    /*----------------------------------------------------------------------*/
    INT_DISABLE();
    time_count = 0;                         /* RESET COUNTER FOR 64BIT COUNT*/
/*
    set_ivtp(DEFAULT);  */                  /* SET IVTP TO .vector SECTION  */
    install_int_vector((void *)c_int45,02); /* SET TINT0 INTERRUPT VECTOR   */
    set_iie(TIMER0);                        /* ENABLE TINT0 BIT IN IIE      */
    INT_ENABLE();                           /* ENABLE GIE BIT               */
}

cputimeExit()
{
}
#endif
#else
#ifdef Sun4sol_TARGET
#include <sys/time.h>

double cputime(option)
enum cputime_option option;
{
	return (gethrvtime()/1e9);
}

cputimeInit()
{
}

cputimeExit()
{
}
#else
#include <sys/time.h>
#include <sys/resource.h>

double cputime(option)
enum cputime_option option;
{
	static struct rusage ru;
	double result;

	getrusage(RUSAGE_SELF,&ru);
 
	result = 0;

	if (option == System || option == All)
	 { result = ru.ru_stime.tv_sec + (ru.ru_stime.tv_usec/1000000.);
	 }
	if (option == User || option == All)
	 { result = ru.ru_utime.tv_sec + (ru.ru_utime.tv_usec/1000000.);
	 }    
	return (result);
}

cputimeInit()
{
}

cputimeExit()
{
}
#endif
#endif
#endif
