#ifdef Tsp_TARGET         

#include <stdio.h>
#include <cdefs.h>        
#include <conc.h>
#include <math.h>
#include "tgen.h"
#include "tgenTsp.h"
#include "servo.h"
#include <vector.h>
#include <puma_kynvar.h>
#include <robotMath.h>
#include "accTrack.h"

#define SEC_TO_TICKS    (1000000/64.)

#if 0
#define DEAD	1.0
#define GAIN	0.06
#define KGAIN	0.0
#define FLIMIT	6.0
#else
#define DEAD	0.0
#define GAIN	0.04
#define KGAIN	0.5
#define FLIMIT	6.0
#endif

#define USE_ACCTRACK	0

typedef struct {
	TRSF Tool;
	TRSF FStoTool;
} ComplyCtrl;

ComplyCtrl compCtrl;
int cnt;
extern int forceFilterWin;
extern int useReducedForce;

MotionContext ctx;

#define RTPRINT	0

double blendFxn(double x, double lim)
{
	double y;

	x /= lim;
	y = (1-exp(-2*x))/(1+exp(-2*x));
	return (y*lim);
}

doComply (tg, arg, ptr)
Tgen *tg;
int arg;
void *ptr;
{
	static int cnt = 0;
	static MotionState state;

	FORCE force;
	ComplyCtrl *cc = ptr;
	TRSF posDel;
	TRSF posErr;
	DIFF velOut;

#if USE_ACCTRACK
	if (cnt == 0)
	 { zeroVect (&state.pos);
	   zeroVect (&state.vel);
	 }
#endif
	cnt++;
	
	multLiTrsf (&posErr, &tg->T6real, &tg->T6);

	transForce (&force, &tg->force, &cc->FStoTool);
	if (ABS(force.f.x) > DEAD)
	 { force.f.x -= SGN(force.f.x)*DEAD;
	 }
	else
	 { force.f.x = 0;
	 }
	if (ABS(force.f.y) > DEAD)
	 { force.f.y -= SGN(force.f.y)*DEAD;
	 }
	else
	 { force.f.y = 0;
	 }
	if (ABS(force.f.z) > DEAD)
	 { force.f.z -= SGN(force.f.z)*DEAD;
	 }
	else
	 { force.f.z = 0;
	 }
#if 1
	velOut.t.x = (force.f.x + KGAIN*posErr.p.x)*GAIN;
	velOut.t.y = (force.f.y + KGAIN*posErr.p.y)*GAIN;
	velOut.t.z = (force.f.z + KGAIN*posErr.p.z)*GAIN;
#else
	velOut.t.x = blendFxn(force.f.x,FLIMIT)*GAIN;
	velOut.t.y = blendFxn(force.f.y,FLIMIT)*GAIN;
	velOut.t.z = blendFxn(force.f.z,FLIMIT)*GAIN;
#endif	
	posDel = cc->Tool;
	multTrsfXyz (&posDel, velOut.t.x, velOut.t.y, velOut.t.z);
	multRiTrsf (&posDel, &posDel, &cc->Tool);
	multTrsf (&tg->Alter, &tg->Alter, &posDel);
#if USE_ACCTRACK
	/* Warning: this will work only for translational motions */
	accTrack (&state, &tg->Alter.p, &state, &ctx);
	tg->Alter.p = state.pos;
#endif

#if RTPRINT
	if ((cnt % 50) == 0)
	 { printf ("%8.3f %8.3f %8.3f\n", force.f.x, force.f.y, force.f.z);
	 }
#endif
}

Tgen *tg;

void exitTask()
{
	getchar();
	tgenRelease (tg);
	exit (1);
}

main()
{
	Process *exitProc;
	JntVec jc;
	TRSF TC, TH;
	TspTgenCtrl *ctrl;
	double interval = 0.005;

	printf ("Allocating trajectory generator ...\n");

	if ((tg = tgenAlloc ("caesar")) == NULL)
	 { printErrors ();
	   exit (1);
	 }
	printf ("done\n");

	forceFilterWin = 2;
	useReducedForce = 1;
	interval = 0.005;

	ctx.rate = 1/interval;
	ctx.velMax = 200.0;
	ctx.accMax = 1000.0;

	printf ("Starting trajectory generator ...\n");
	if (tgenStart (tg, interval, LINK3OUT, LINK3IN, LINK2OUT, LINK2IN) < 0)
	 { printErrors();
	   exit (1);
	 }
	printf ("done\n");

	ctrl = (TspTgenCtrl*)tg->ctrl;

	ctrl->resetBias = 1;
	tgenSetPath (tg, TGEN_CART_PATH);
	tgenMoveC (tg, &tg->T6, (TRSF*)0);
	tgenStop (tg, 100000.0);

	identTrsf (&compCtrl.Tool);
	identTrsf (&compCtrl.FStoTool);
	xyzToTrsf (&compCtrl.Tool, 0, 0, 191.4);
	xyzToTrsf (&compCtrl.FStoTool, 0, 0, 90.0);

	ctrl->resetBias = 1;
	ProcWait (1*SEC_TO_TICKS);

	printf ("Starting compliance: (<CR> to confirm)\n");
	getchar();

	tgenAlterFxnSet (tg, doComply, 0, &compCtrl);

#if !RTPRINT
	exitProc = ProcAlloc (exitTask, 0x2000, /*num params=*/0);
	ProcRun (exitProc);		
	printf ("Hit <CR> to exit\n");
	while (1)
	 { extern int controlFxnCnt;
	   extern int controlFxnTime;

	   ProcWait (1*SEC_TO_TICKS);
	   printf ("time = %g usec\n", 
		    controlFxnTime/(double)controlFxnCnt);
	 }
#else
	while (1)
	 { ProcWait (1*SEC_TO_TICKS);
	 }
#endif
}

#endif
