/* kin.c */

#ifdef Tsp_TARGET

#include <conc.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "xkin.h"
#include "gripper.h"

static Channel *ToKin, *FromKin;

#define PI 3.14159265358979323846

int KinMsgLenOut[KIN_FN] =
{       0,
        DEPART_MSG_OUT,
        WTOJ_MSG_OUT,
        OL_WTOJ_MSG_OUT,
        OL_WMAT_TOJ_MSG_OUT,
        JTOW_MSG_OUT,
        OL_DEPART_MSG_OUT,
        CFG_MSG_OUT,
        KIN_TST_MSG_OUT,
        MTOW_MSG_OUT,
        WTOM_MSG_OUT,
        FORCECON_MSG_OUT,
	0,			/* FORCESET */
	B430_MSG_OUT,
        0       };

int KinMsgLenIn[KIN_FN] =
{       0,
        DEPART_MSG_IN,
        WTOJ_MSG_IN,
	OL_WTOJ_MSG_IN,
        OL_WMAT_TOJ_MSG_IN,
        JTOW_MSG_IN,
        OL_DEPART_MSG_IN,
        CFG_MSG_IN,
	12,
/*
        KIN_TST_MSG_IN,
*/
        MTOW_MSG_IN,
        WTOM_MSG_IN,
        FORCECON_MSG_IN,
	0,			/* FORCESET */
	B430_MSG_IN,
        0   };

int Task(buf)
int *buf;
{
  char TaskIdReg, status, l;
  int f;
 
  f = buf[0];

#if 0
printf( "task: cmd=start type=%d len=%d ", f, KinMsgLenIn[f] );
fflush(stdout);
#endif
  ChanOutChar(ToKin, IF_TASK_START);
  ChanOutChar(ToKin, TaskIdReg);
  ChanOutInt(ToKin, KinMsgLenIn[f]);
  ChanOut(ToKin, buf, KinMsgLenIn[f]);
  status = ChanInChar(FromKin); 
#if 0
printf( "status=%d\n", status );
#endif
  if (status != 0)
    return -1;
#if 0
printf( "task: cmd=output len=%d ", KinMsgLenOut[f] );
fflush(stdout);
#endif
  ChanOutChar(ToKin, IF_TASK_OUTPUT);
  ChanOutChar(ToKin, TaskIdReg);
  l = ChanInInt(FromKin);
  ChanIn(FromKin, buf, KinMsgLenOut[f]);
  status = ChanInChar(FromKin);
#if 0
printf( "status=%d\n", status );
#endif
  if (status != 0)
    return -1;

  return 0;
}

int buf[256/4];

#define GRIPPER_MAX_ADC	    122

static int GenGripperCmd( int type, int val, float *posn, float *feedback )
{
    int i;

    GRIP_REQ *req = (GRIP_REQ*)&buf[1];
    GRIP_REP *rep = (GRIP_REP*)&buf[2];

    buf[0] = KIN_B430_FN;
    req->type = type;
    req->gripper.force = val;
    
    if( Task(buf) < 0 )
	printf("Bad Task\n");

    if( posn )
	*posn = rep->gripper_posn * (2.0/GRIPPER_MAX_ADC);
    if( feedback )
	for( i = 0; i < 6; i++ )
	    feedback[i] = rep->force_feedback[i] / 512.0;

    return( buf[1] );
}

int GripperForce( double force, float *oldposn )
{
    int iforce = 100 * force;
    if( iforce < -100 || iforce > +100 )
	return( -1 );
    return( GenGripperCmd( GRIPPER_FORCE, iforce, oldposn, 0 ) );
}

int GripperPosn( double posn, float *oldposn )
{
    int iposn = (GRIPPER_MAX_ADC/2.0) * posn;
    if( iposn < 0 || iposn > GRIPPER_MAX_ADC )
	return( -1 );
    return( GenGripperCmd( GRIPPER_POSN, iposn, oldposn, 0 ) );
}

int GripperQuery( float *oldposn, float *feedback )
{
    return( GenGripperCmd( GRIPPER_NOOP, 0, oldposn, feedback ) );
}

forceSetChan (to, from)
Channel *to, *from;
{
	ToKin = to;
	FromKin = from;
}

int forceGetTime;
int forceFilterWin = 5;

int forceGet (force)
float *force;
{
	int i;
#if 0
	return ( GenGripperCmd( GRIPPER_NOOP, 0, (float*)0, force ) );
#else
	int forceData[6];

	ChanOutChar (ToKin, forceFilterWin);
	ChanIn (FromKin, forceData, 6*sizeof(int));
	forceGetTime = (unsigned char)ChanInChar(FromKin);
	for (i=0; i<6; i++)
	 { force[i] = forceData[i]/512.;
	 }
#endif
}

#endif
