/* kin.c */

#ifdef Tsp_TARGET

#include <conc.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "xkin.h"
#include "gripper.h"

#define FREE_ALL_BIT                    0x80
#define FREE_WRIST_BIT                  0x40
#define FREE_ELBOW_BIT                  0x20
#define FREE_REACH_BIT                  0x10

#define WRIST_BIT                               0x08
#define ELBOW_BIT                               0x02
#define REACH_BIT                               0x01
#define CFG_UPRIGHT                             0x04    /* Robot Upright config */

extern Channel *ToKin, *FromKin;

#define PI 3.14159265358979323846

float Tr_n[6] = {
  1.0 / ( 100.0 / (2 * PI) * 1000.0),
  1.0 / ( 100.0 / (2 * PI) * 1000.0),
  1.0 / (-100.0 / (2 * PI) * 1000.0),
  1.0 / (-101.0 / (2 * PI) * 500.0),
  1.0 / (-100.0 / (2 * PI) * 500.0),
  1.0 / (-100.0 / (2 * PI) * 500.0),
};

int MotorToJoint(motor, joint)
int motor[8];
float joint[8];
{
   joint[0] = Tr_n[0] * (float) motor[0];
   joint[1] = Tr_n[1] * (float) motor[1];
   joint[2] = Tr_n[2] * (float) motor[2] - joint[1] / (-100.0);
   joint[3] = Tr_n[3] * (float) motor[3];
   joint[4] = Tr_n[4] * (float) motor[4];
   joint[5] = Tr_n[5] * (float) motor[5];
}

int JointToMotor(joint, motor)
float joint[8];
int motor[8];
{
   motor[0] = (int) (joint[0] / Tr_n[0]);
   motor[1] = (int) (joint[1] / Tr_n[1]);
   motor[2] = (int) ((joint[2] + joint[1] / (-100.0))/ Tr_n[2]);
   motor[3] = (int) (joint[3] / Tr_n[3]);
   motor[4] = (int) (joint[4] / Tr_n[4]);
   motor[5] = (int) (joint[5] / Tr_n[5]);
}


int KinMsgLenOut[KIN_FN] =
{       0,
        DEPART_MSG_OUT,
        WTOJ_MSG_OUT,
        OL_WTOJ_MSG_OUT,
        OL_WMAT_TOJ_MSG_OUT,
        JTOW_MSG_OUT,
        OL_DEPART_MSG_OUT,
        CFG_MSG_OUT,
        KIN_TST_MSG_OUT,
        MTOW_MSG_OUT,
        WTOM_MSG_OUT,
        FORCECON_MSG_OUT,
	0,			/* FORCESET */
	B430_MSG_OUT,
        0       };

int KinMsgLenIn[KIN_FN] =
{       0,
        DEPART_MSG_IN,
        WTOJ_MSG_IN,
        OL_WTOJ_MSG_IN,
        OL_WMAT_TOJ_MSG_IN,
        JTOW_MSG_IN,
        OL_DEPART_MSG_IN,
        CFG_MSG_IN,
	12,
/*
        KIN_TST_MSG_IN,
*/
        MTOW_MSG_IN,
        WTOM_MSG_IN,
        FORCECON_MSG_IN,
	0,			/* FORCESET */
	B430_MSG_IN,
        0   };

int Task(buf)
int *buf;
{
  char TaskIdReg, status, l;
  int f;
 
  f = buf[0];

#if 0
printf( "task: cmd=start type=%d len=%d ", f, KinMsgLenIn[f] );
fflush(stdout);
#endif
  ChanOutChar(ToKin, IF_TASK_START);
  ChanOutChar(ToKin, TaskIdReg);
  ChanOutInt(ToKin, KinMsgLenIn[f]);
  ChanOut(ToKin, buf, KinMsgLenIn[f]);
  status = ChanInChar(FromKin); 
#if 0
printf( "status=%d\n", status );
#endif
  if (status != 0)
    return -1;
#if 0
printf( "task: cmd=output len=%d ", KinMsgLenOut[f] );
fflush(stdout);
#endif
  ChanOutChar(ToKin, IF_TASK_OUTPUT);
  ChanOutChar(ToKin, TaskIdReg);
  l = ChanInInt(FromKin);
  ChanIn(FromKin, buf, KinMsgLenOut[f]);
  status = ChanInChar(FromKin);
#if 0
printf( "status=%d\n", status );
#endif
  if (status != 0)
    return -1;

  return 0;
}

int buf[256/4];

#define GRIPPER_MAX_ADC	    122

static int GenGripperCmd( int type, int val, float *posn, float *feedback )
{
    int i;

    GRIP_REQ *req = (GRIP_REQ*)&buf[1];
    GRIP_REP *rep = (GRIP_REP*)&buf[2];

    buf[0] = KIN_B430_FN;
    req->type = type;
    req->gripper.force = val;
    
    if( Task(buf) < 0 )
	printf("Bad Task\n");

    if( posn )
	*posn = rep->gripper_posn * (2.0/GRIPPER_MAX_ADC);
    if( feedback )
	for( i = 0; i < 6; i++ )
	    feedback[i] = rep->force_feedback[i] / 512.0;

    return( buf[1] );
}

int GripperForce( double force, float *oldposn )
{
    int iforce = 100 * force;
    if( iforce < -100 || iforce > +100 )
	return( -1 );
    return( GenGripperCmd( GRIPPER_FORCE, iforce, oldposn, 0 ) );
}

int GripperPosn( double posn, float *oldposn )
{
    int iposn = (GRIPPER_MAX_ADC/2.0) * posn;
    if( iposn < 0 || iposn > GRIPPER_MAX_ADC )
	return( -1 );
    return( GenGripperCmd( GRIPPER_POSN, iposn, oldposn, 0 ) );
}

int GripperQuery( float *oldposn, float *feedback )
{
    return( GenGripperCmd( GRIPPER_NOOP, 0, oldposn, feedback ) );
}

int WorldToJoint(config, world, oldjoint, joint)
unsigned char config;
float world[8], oldjoint[8], joint[8];
{
  int i;

  buf[0] = KIN_WTOJ_FN;
  buf[1] = config;
  bcopy(world, &buf[2], 32);
  bcopy(oldjoint, &buf[10], 32);
  if (Task(buf) < 0)
    printf("Bad Task\n");;
  bcopy(&buf[2], joint, 32);
  return buf[1]; 
}

int WorldToMotor(config, world, oldjoint, motor)
unsigned char config;
float world[8], oldjoint[8];
int motor[8];
{
  int i;

  buf[0] = KIN_WTOM_FN;
  buf[1] = config;
  bcopy(world, &buf[2], 32);
  bcopy(oldjoint, &buf[10], 32);
  if (Task(buf) < 0)
    printf("Bad Task\n");;
  bcopy(&buf[2], motor, 32);
  return buf[1]; 
}

int Config(config, joint, newconfig)
unsigned char config;
float joint[8];
unsigned char *newconfig;
{
  buf[0] = KIN_CFG_FN;
  buf[1] = config;
  bcopy(joint, &buf[2], 32);
  if (Task(buf) < 0)
    printf("Bad Task\n");;
  *newconfig = buf[2];
  return buf[1];
}


int Test(i)
int i;
{
  buf[0] = KIN_TST;
  buf[2] = i;
  printf("In %d\n", buf[2]);
  if (Task(buf) < 0)
    printf("Bad Task\n");;
  printf("Out %d\n", buf[2]);
  return buf[2];
}

#endif
