/* motion.c */

#ifdef Tsp_TARGET

#include <conc.h>
#include <stdio.h>

extern Channel *ToServo, *FromServo, *ToKin, *FromKin;

#define LIMIT(x, A)  {if ((x) > (A)) (x) = (A); if ((x) < -(A)) (x) = -(A);}

int trapezoidal(setpoint, vel, maxv, maxa, goal)
int *setpoint, *vel, maxv, maxa, goal;
{
  int maxslow, a;

  if (*vel >= 0) {
    if (*vel > maxv)
      a = -maxa;
    else
    for (a = maxa; a > -maxa ; a--)
      if (goal - (*setpoint + *vel + a) >= ((*vel+a)*(*vel+a) - (*vel+a)*maxa + maxa) / (2 * maxa)) {
        break;
      }
  }
  else {
    if (*vel < -maxv)
      a = maxa;
    else
    for (a = -maxa; a < maxa ; a++)
      if (goal - (*setpoint + *vel + a) <= -((*vel+a)*(*vel+a) + (*vel+a)*maxa + maxa) / (2 * maxa)) {
        break;
      }
  }

  *vel += a;
  LIMIT(*vel, maxv);
  *setpoint += *vel;

  return *vel == 0 && *setpoint == goal;

}

#define MAXV 30
#define MAXA 1 

static struct {int vel, maxv, maxa; } v[8] = {
  {0, MAXV, MAXA},
  {0, MAXV, MAXA},
  {0, MAXV, MAXA},
  {0, MAXV, MAXA},
  {0, MAXV, MAXA},
  {0, MAXV, MAXA},
  {0, MAXV, MAXA},
  {0, MAXV, MAXA},
};

static int gposition[8], gsetpoint[8];

static int gstatus[8] = {
  0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x08, 0x08
};

int ggoal[8], complete;

int motion()
{
  int i, t;


  if (Interpolate(10) < 0)
    return -1;

  if (Listen(gstatus, gsetpoint, gposition) < 0)
    return -1;

/*
  printf("Setpoint: %d %d %d %d %d %d %d %d\n", gsetpoint[0], gsetpoint[1],
   gsetpoint[2], gsetpoint[3], gsetpoint[4], gsetpoint[5], gsetpoint[6], 
   gsetpoint[7]);

  printf("Position: %d %d %d %d %d %d %d %d\n", gposition[0], gposition[1],
   gposition[2], gposition[3], gposition[4], gposition[5], gposition[6], 
   gposition[7]);

  printf("Goal: %d %d %d %d %d %d %d %d\n", ggoal[0], ggoal[1],
   ggoal[2], ggoal[3], ggoal[4], ggoal[5], ggoal[6], ggoal[7]);

  printf("Status: %x %x %x %x %x %x %x %x\n", gstatus[0], gstatus[1],
   gstatus[2], gstatus[3], gstatus[4], gstatus[5], gstatus[6], gstatus[7]);
*/

  ProcToHigh();
  t = Time();
  for (;;) {
    complete = 1;
    for (i = 0; i < 8; i++)
      complete = 
        trapezoidal(&gsetpoint[i], &v[i].vel, v[i].maxv, v[i].maxa, ggoal[i])
        && complete;

    if (Normal(gstatus, gsetpoint, gposition) < 0)
      return -1;
/*
  printf("Setpoint: %d %d %d %d %d %d %d %d\n", gsetpoint[0], gsetpoint[1],
   gsetpoint[2], gsetpoint[3], gsetpoint[4], gsetpoint[5], gsetpoint[6], 
   gsetpoint[7]);
*/

    t+= 10000;
    ProcAfter(t);  /* 64 * 156 = 10ms */
    if (complete)
      break;
  }
  ProcToLow();
  
}
#endif /* Tsp_TARGET */

