#include <math.h>
#include <robotMath.h>
#include <robot_kynvar.h>
#include <vector.h>
#include "tgen.h"

#define JACOB_EPSILON	(1e-5)
#define CNT_EPS	0.001

int timeCnt PROTO((double, int));

int timeCnt (t, min)
double t;
int min;
{
	return (MAX(ceil(t-CNT_EPS), min));
}

/*OK*/double alphaFxn(s)
double s;
{
	return (((6*s - 15)*s + 10)*s*s*s);
}

/*OK*/double betaFxn(s)
double s;
{
	return ((((s - 3)*s + 3)*s - 1)*s*s*s);
}

/*OK*/motionControlInit (mc1, mc0, tg)
MotionControl *mc1, *mc0;
Tgen *tg;
{
	if (mc1->req->pathType == TGEN_CART_PATH)
	 { motionControlInitCart (mc1, mc0, tg);
	 }
	else /* mc1->req->pathType == TGEN_JNT_PATH */
	 { motionControlInitJnt (mc1, mc0, tg);
	 }
	mc1->segCntEst = mc1->segCnt;
}

motionControlInitCart (mc1, mc0, tg)
MotionControl *mc1, *mc0;
Tgen *tg;
{
	MotionRequest *req0, *req1;
	DIFF cvel0, cvel1;

	req1 = mc1->req;
	req0 = req1->prev;

	if (req1->flags & TGEN_STOP)
	 { zeroDiff (&cvel1);
	   mc1->segCnt = timeCnt(req1->segTime/tg->sampInterval, 1);
	 }
	else
	 { TRSF TCinv;
	   double segTime, minTime;

	   segTime = req1->segTime;
	   minTime = 0;

	   /* compute a maximum time only if the transition time is to
	      be computed automatically. */
	   if (req1->segTime == -1 || req1->transTime == -1)
	    { double d;

	      if (req1->segTime == -1)
	       { segTime = 0;
	       }
	      if ((d = normVect (&req1->cdrive.p)) != 0.0)
	       { if (req1->segTime == -1)
		  { segTime = MAX(segTime, d/req1->maxTransVel);
		  }
		 if (req1->transTime == -1)
		  { minTime = MAX(minTime, sqrt(2*d/req1->maxTransAcc));
		  }
	       }
	      if ((d = ABS (req1->cdrive.a)) != 0.0)
	       { if (req1->segTime == -1)
		  { segTime = MAX(segTime, d/req1->maxRotVel);
		  }
		 if (req1->transTime == -1)
		  { minTime = MAX(minTime, sqrt(2*d/req1->maxRotAcc));
		  }
	       }
	    }
	   segTime = MAX(minTime, segTime);
	   mc1->segCnt = timeCnt(segTime/tg->sampInterval, 1);
	   segTime = mc1->segCnt*tg->sampInterval;

	   /* now compute the velocity, in TC base coordinates. This is
	      done by dividing cdrive by segTime, *and* negating it,
	      since cdrive is directed from TC to the old TC. */

	   scaleVect (&cvel1.t, -1/segTime, &req1->cdrive.p);
	   scaleVect (&cvel1.r, -req1->cdrive.a/segTime, &req1->cdrive.u);

	   invertTrsf (&TCinv, &req1->TC);
	   rotateDiff (&cvel1, &cvel1, &TCinv);
	 }
	mc1->cvel = cvel1;	/* save for later */

	/* determine previous velocity */

	if (mc0 == NULL || req0->flags & TGEN_STOP)
	 { zeroDiff (&cvel0);
	 }
	else
	 { TRSF TBinv;

	   invertTrsf (&TBinv, &req1->TB);

	   if (req0->pathType == TGEN_CART_PATH)
	    { TRSF ToolChg;

	      rotateDiff (&cvel0, &mc0->cvel, &req0->TC);
	      multLiTrsf (&ToolChg, &req0->TL, &req1->TL);
	      transDiff (&cvel0, &cvel0, &ToolChg);
#if 0
	      if (req1->id == 17)
	       { printVf ("cvel:   %10.6f\n", (float*)&mc0->cvel, 6);
		 printVf ("cvelP:  %10.6f\n", (float*)&cvel0, 6);
		 printTrsf ("T:      %m\n", &ToolChg);
	       }
#endif
	    }
	   else /* TGEN_JNT_PATH */
	    { fwdJacob (&cvel0, mc0->jvel.v, req0->jc.v, tg->kyn);
	      transDiff (&cvel0, &cvel0, &req1->TL); 
	    }
	   rotateDiff (&cvel0, &cvel0, &TBinv);
	 }

	/* compute change in velocity and then transition time */

	 { double transTime;

	   subtractVect (&mc1->delCvel.t, &cvel1.t, &cvel0.t);
	   subtractVect (&mc1->delCvel.r, &cvel1.r, &cvel0.r);
	   mc1->prevCvel = cvel0;
	   transTime = mc1->req->transTime;
	   if (transTime == -1)
	    { transTime = MAX(normVect(&mc1->delCvel.t)/req1->maxTransAcc,
			      normVect(&mc1->delCvel.r)/req1->maxRotAcc);
	      mc1->transCnt = timeCnt(transTime/(tg->sampInterval*2), 1);
	    }
	   else
	    { mc1->transCnt = timeCnt(transTime/(tg->sampInterval*2), 0);
	    }
	 };
}

motionControlInitJnt (mc1, mc0, tg)
MotionControl *mc1, *mc0;
Tgen *tg;
{
	MotionRequest *req0, *req1;
	JntVec jvel0;

	req1 = mc1->req;
	req0 = req1->prev;

	if (req1->flags & TGEN_STOP)
	 { zeroVf (mc1->jvel.v, NJNTS);
	   mc1->segCnt = timeCnt(req1->segTime/tg->sampInterval, 1);
	 }
	else
	 { double segTime, minTime;

	   segTime = req1->segTime;
	   minTime = 0;

	   if (req1->segTime == -1 || req1->transTime == -1)
	    { /* compute segment time */

	      double d;
	      int j;

	      if (req1->segTime == -1)
	       { segTime = 0;
	       }
	      for (j=0; j<NJNTS; j++)
	       { if ((d = ABS(req1->jdrive.v[j])) != 0.0)
		  { if (req1->segTime == -1)
		     { segTime = MAX(segTime, d/req1->maxJntVel.v[j]);
		     }
		    if (req1->transTime == -1)
		     { minTime = MAX(minTime, sqrt(2*d/req1->maxJntAcc.v[j]));
		     }
		  }
	       }
	    }
	   segTime = MAX(minTime, segTime);
	   mc1->segCnt = timeCnt(segTime/tg->sampInterval, 1);
	   segTime = mc1->segCnt*tg->sampInterval;

	   /* now compute the velocity */

	   scaleVf (mc1->jvel.v, 1/segTime, req1->jdrive.v, NJNTS);
	 }

	/* determine previous velocity */

	if (mc0 == NULL || req0->flags & TGEN_STOP)
	 { zeroVf (jvel0.v, NJNTS);
	 }
	else if (req0->pathType == TGEN_JNT_PATH)
	 { jvel0 = mc0->jvel;
	 }
	else /* TGEN_CART_PATH */
	 { DIFF cvel0;
	   TRSF TLinv;

	   rotateDiff (&cvel0, &mc0->cvel, &req0->TC);
	   invertTrsf (&TLinv, &req0->TL);
	   transDiff (&cvel0, &cvel0, &TLinv);
	   invJacob (jvel0.v, &cvel0, req0->jc.v, tg->kyn, JACOB_EPSILON);
	 }

	/* compute change in velocity and then transition time */

	 { double transTime;

	   subtractVf (mc1->delJvel.v, mc1->jvel.v, jvel0.v, NJNTS);
	   mc1->prevJvel = jvel0;
	   transTime = mc1->req->transTime;
	   if (transTime == -1)
	    { int j;

	      transTime = 0;
	      for (j=0; j<NJNTS; j++)
	       { transTime = MAX(transTime, ABS(mc1->delJvel.v[j])/
					    req1->maxJntAcc.v[j]);
	       }
	      mc1->transCnt = timeCnt(transTime/(tg->sampInterval*2), 1);
	    }
	   else
	    { mc1->transCnt = timeCnt(transTime/(tg->sampInterval*2), 0);
	    }
	 };
}

/*OK*/void computeLinearMotionCart (TN, TC, cdrive, sc)
TRSF *TN, *TC;
DSPL *cdrive;
double sc;
{
	DSPL cdriveScaled;
	TRSF Td;

	scaleDspl (&cdriveScaled, 1.0-sc, cdrive);
	dsplToTrsf (&Td, &cdriveScaled);
	multTrsf (TN, TC, &Td);
}

/*OK*/void computeLinearMotionJnt (jn, jc, jdrive, sc)
JntVec *jn, *jc;
JntVec *jdrive;
double sc;
{
	JntVec jdriveScaled;

	scaleVf (jdriveScaled.v, 1.0-sc, jdrive->v, NJNTS);
	subtractVf (jn->v, jc->v, jdriveScaled.v, NJNTS);
}

computeT6Cart (T6, mctl)
TRSF *T6;
MotionControl *mctl;
{
	TRSF T1;

	computeLinearMotionCart (&T1, &mctl->req->TC,
				      &mctl->req->cdrive, mctl->scale);
	multRiTrsf (T6, &T1, &mctl->req->TL);
}

/*OK*/computeCruiseMotion (tg)
Tgen *tg;
{
	MotionControl *curr;

	curr = tg->currMot;
	if (curr->req->pathType == TGEN_CART_PATH)
	 { computeT6Cart (&tg->T6, curr);
	   tg->conf = curr->req->conf;
	 }
	else /* TGEN_JNT_PATH */
	 { computeLinearMotionJnt (&tg->j6, &curr->req->jc,
				   &curr->req->jdrive, curr->scale);
	 }
}

computeBlendedMotion (tg)
Tgen *tg;
{
	MotionControl *curr;
	MotionControl *prev;

	double sb;
	double alpha, bscale;

	curr = tg->currMot;
	prev = tg->prevMot;

	sb = (curr->cnt + curr->transCnt)/(double)(2*curr->transCnt);

	alpha = alphaFxn(sb);
	bscale = 2*curr->transCnt*tg->sampInterval*6*betaFxn(sb);

	if (curr->req->pathType == TGEN_CART_PATH)
	 { TRSF T1, T2, T1C, TC2, TA, TB, TT;
	   VECT ptemp, pc, u1c, uc2;
	   float ang1c, angc2;

	   computeLinearMotionCart (&T2, &curr->req->TC, 
				    &curr->req->cdrive, curr->scale);
	   if (prev->req->pathType == TGEN_CART_PATH)
	    { computeT6Cart (&T1, prev);
	    }
	   else /* TGEN_JNT_PATH */
	    {   
	      if (tg->flags & TGEN_ROBOT_ATTACHED)
	       { JntVec j1;

		 if (prev->req->flags & TGEN_STOP)
		  { j1 = prev->req->jc;
		  }
		 else
		  { computeLinearMotionJnt (&j1, &prev->req->jc,
					    &prev->req->jdrive, prev->scale);
		  }
		 fwdKinematics (&T1,(unsigned long *)0,j1.v,tg->kyn,(void*)0);
	       }
	      else
	       { if (prev->req->flags & TGEN_STOP)
		  { multLiTrsf (&T1, &prev->req->TC, &prev->req->TL);
		  }
		 else
		  { tgenAbort ("Panic: stop motion not inserted\n");
		    return (-1);
		  }
	       }
	    }
	   multTrsf (&T1, &T1, &curr->req->TL);
	   combineVect (&ptemp, alpha, &T2.p, 1-alpha, &T1.p);
	   scaleVect (&pc, -bscale, &curr->delCvel.t);
	   addVect (&ptemp, &pc, &ptemp);

	   /* now hack the rotation part */

	   multLiTrsf (&T1C, &T1, &curr->req->TB);
	   multLiTrsf (&TC2, &curr->req->TB, &T2);
	   trsfToRot (&u1c, &ang1c, &T1C);
	   trsfToRot (&uc2, &angc2, &TC2);
	   rotToTrsf (&TA, &u1c, alpha*ang1c + 
				 RADTODEG*bscale*normVect(&curr->prevCvel.r));
	   rotToTrsf (&TB, &uc2, alpha*angc2 - 
				 RADTODEG*bscale*normVect(&curr->cvel.r));
	   multTrsf (&TT, &T1, &TA);
	   multTrsf (&TT, &TT, &TB);
	   TT.p = ptemp;
	   multRiTrsf (&tg->T6, &TT, &curr->req->TL);
#if 0
	   if (curr->req->id == 17 && curr->cnt == -2)
	    { printTrsf ("T1: %10.5m\n", &T1);
	      printTrsf ("T2: %10.5m\n", &T2);
	      printTrsf ("TB: %10.5m\n", &curr->req->TB);
	      printVf ("VP: %10.5f\n", (float*)&curr->prevCvel, 6);
	      printVf ("VC: %10.5f\n", (float*)&curr->cvel, 6);
	      printTrsf ("TA: %10.5m\n", &TA);
	      printTrsf ("TB: %10.5m\n", &TB);
	    }
#endif
	   tg->conf = curr->req->conf;
	 }
	else /* TGEN_JNT_PATH */
	 { JntVec j1, j2, jc;

	   computeLinearMotionJnt (&j2, &curr->req->jc, 
				   &curr->req->jdrive, curr->scale);
	   if (prev->req->pathType == TGEN_JNT_PATH)
	    { computeLinearMotionJnt (&j1, &prev->req->jc,
				      &prev->req->jdrive, prev->scale);
	    }
	   else /* TGEN_CART_PATH */
	    { if (prev->req->flags & TGEN_STOP)
	       { j1 = prev->req->jc;
	       }
	      else
	       { TRSF T1;
		 computeT6Cart (&T1, prev);
		 invKinematics (j1.v, &T1, prev->req->conf, tg->j6.v, 
			   	      tg->kyn, tg->jls, (void*)0);
		 /* FIX XXX status return check */
	       }
	    }
	   combineVf (tg->j6.v, alpha, j2.v, 1-alpha, j1.v, NJNTS);
	   scaleVf (jc.v, -bscale, curr->delJvel.v, NJNTS);
	   addVf (tg->j6.v, jc.v, tg->j6.v, NJNTS);
#if 0
	   if (curr->req->id == 18 && curr->cnt == -19)
	    { printJf ("j1:      %10.5f\n", j1.v, NJNTS);
	      printJf ("j2:      %10.5f\n", j2.v, NJNTS);
	      printJf ("delJvel: %10.5f\n", curr->delJvel.v, NJNTS);
	      printJf ("jvel:    %10.5f\n", curr->jvel.v, NJNTS);
	      printJf ("prevJvel:%10.5f\n", curr->prevJvel.v, NJNTS);
	    }
#endif
	 }
}

currentMotionInit (tg)
Tgen *tg;
{
	MotionControl *curr;
	MotionControl *prev;
	double fudge;
	int redoDelVel = 0;

	curr = tg->currMot;
	prev = tg->prevMot;

	curr->cnt = -curr->transCnt;
	curr->scale = computeScale(curr);
	curr->status = TGEN_ACTIVE;
	if (tg->ctrl == NULL)
	 { printf ("#%s: segCnt=%d transCnt=%d\n", 
		     (curr->req->flags & TGEN_STOP) ? 
		     ((curr->req->flags & 1) ? "IDLE" : "STOP") :
	     (curr->req->targType == TGEN_CART_TARG) ? "MOVE C" : 
		     "MOVE J",
		     curr->segCnt,  
		     curr->transCnt);
	 }
	/* patch up velocity info, if necessary */

	if (curr->req->pathType == TGEN_CART_PATH)
	 { if (prev->segCnt != prev->segCntEst)
	    { fudge = prev->segCntEst/(double)prev->segCnt;
	      scaleVf ((float*)&curr->prevCvel, fudge, 
		       (float*)&curr->prevCvel, 6);
	      redoDelVel = 1;
	    }
	   if (curr->segCnt != curr->segCntEst)
	    { fudge = curr->segCntEst/(double)curr->segCnt;
	      scaleVf ((float*)&curr->cvel, fudge, 
		       (float*)&curr->cvel, 6);
	      redoDelVel = 1;
	    }
	   if (redoDelVel)
	    { subtractVf ((float*)&curr->delCvel, (float*)&curr->cvel,
			  (float*)&curr->prevCvel, 6);
	    }
	 }
	else /* TGEN_JNT_PATH */
	 { if (prev->segCnt != prev->segCntEst)
	    { fudge = prev->segCntEst/(double)prev->segCnt;
	      scaleVf (curr->prevJvel.v, fudge, 
		       curr->prevJvel.v, NJNTS);
	      redoDelVel = 1;
	    }
	   if (curr->segCnt != curr->segCntEst)
	    { fudge = curr->segCntEst/(double)curr->segCnt;
	      scaleVf (curr->jvel.v, fudge, 
		       curr->jvel.v, NJNTS);
	      redoDelVel = 1;
	    }
	   if (redoDelVel)
	    { subtractVf (curr->delJvel.v, curr->jvel.v,
			  curr->prevJvel.v, NJNTS);
	    }
	 }
}

