#ifndef _PMCREG_H_
#define _PMCREG_H_

#ifdef KERNEL
#include "sys/poll.h"
#else
#include <sys/poll.h>
#endif
#include <sys/segment.h>	/* for ATIIOC_ENABLE_IOPORTS */
#include <enableIO.h>

/* Offsets for the XFER and CTRL registers in ISA control space */

#define PMC_XFER	0
#define PMC_CTRL	2

/* Device registers (accessed using the CTRL and XFER words
   in the ISA I/O space) */

#define PMC_PIT		0x00
#define PMC_PITL	0x00
#define PMC_PITH	0x01
#define PMC_WD		0x02
#define PMC_GATE	0x04
#define PMC_WDCTRL	0x06
#define PMC_SYNCCTRL	0x06

#define PMC_DAC0	0x08
#define PMC_DAC1	0x09
#define PMC_DAC2	0x0a
#define PMC_DAC3	0x0b
#define PMC_DAC4	0x0c
#define PMC_DAC5	0x0d
#define PMC_DAC6	0x0e
#define PMC_DAC7	0x0f
#define PMC_DAC8	0x10
#define PMC_DAC9	0x11
#define PMC_DAC10	0x12
#define PMC_DAC11	0x13
#define PMC_DAC12	0x14
#define PMC_DAC13	0x15
#define PMC_DAC14	0x16
#define PMC_DAC15	0x17

#define PMC_ADC0	0x18
#define PMC_ADC1	0x19
#define PMC_ADC2	0x1a
#define PMC_ADC3	0x1b
#define PMC_ADC4	0x1c
#define PMC_ADC5	0x1d
#define PMC_ADC6	0x1e
#define PMC_ADC7	0x1f
                            
#define PMC_DIGIN0	0x20
#define PMC_DIGIN0L	0x20
#define PMC_DIGIN0H	0x21
#define PMC_DIGIN1	0x22
#define PMC_DIGIN1L	0x22
#define PMC_DIGIN1H	0x23
#define PMC_DIGIN2	0x24
#define PMC_DIGIN2L	0x24
#define PMC_DIGIN2H	0x25
#define PMC_DIGOUT0	0x26
#define PMC_DIGOUT0L	0x26
#define PMC_DIGOUT0H	0x27

#define PMC_DMASK0	0x28
#define PMC_DMASK0L	0x28
#define PMC_DMASK0H	0x29
#define PMC_DMASK1	0x2a
#define PMC_DMASK1L	0x2a
#define PMC_DMASK1H	0x2b
#define PMC_DMASK2	0x2c
#define PMC_DMASK3	0x2d
#define PMC_DACT	0x2e
#define PMC_DACTL	0x2e
#define PMC_DACTH	0x2f

#define PMC_DMFLG	0x30
#define PMC_OEDIG	0x32
#define PMC_INTFLGHOST	0x34
#define	PMC_INTENAHOST	0x35
#define PMC_INTFLGDSP	0x36
#define PMC_INTENADSP	0x37

#define PMC_PIT_INTENT	0x01
#define PMC_SYNC_INTENT 0x02
#define PMC_DMASK0_INTENT 0x04
#define PMC_DMASK1_INTENT 0x08
#define PMC_DMASK2_INTENT 0x10
#define PMC_DMASK3_INTENT 0x20

#define PMC_DSPRESET	0x38
#define PMC_SOFTSYNC	0x3e

/* Quadrature Decoders */
#define PMC_QDCOUNT	0x40
#define PMC_QDCOUNT0	0x40
#define PMC_QDCOUNT1	0x42
#define PMC_QDCOUNT2	0x44
#define PMC_QDCOUNT3	0x46
#define PMC_QDCOUNT4	0x48
#define PMC_QDCOUNT5	0x4a
#define PMC_QDCOUNT6	0x4c
#define PMC_QDCOUNT7	0x4e

#define PMC_QDMODE	0x50
#define PMC_QDMODE0	0x50
#define PMC_QDMODE1	0x52
#define PMC_QDMODE2	0x54
#define PMC_QDMODE3	0x56
#define PMC_QDMODE4	0x58
#define PMC_QDMODE5	0x5a
#define PMC_QDMODE6	0x5c
#define PMC_QDMODE7	0x5e

/* Other useful values */
#define PMC_AUTOINCR	0x80
#define PMC_DSPREG		0x100
#define PMC_ADCBUSY		0x8000
	/* Beginning of internal memory */
#define PMC_INTADDR		0x20000L
	/* Beginning of external memory */
#define PMC_EXTADDR		0x400000L

#define PMC_NUM_CHANS	8	/* number of control channels */
typedef struct {
	dev_info_t *dip;
	int instance;
	int hasHighLevelIntr;

	ddi_iblock_cookie_t iblkCookie;
	ddi_idevice_cookie_t idevCookie;
/*	ddi_acc_handle_t regsHandle; */
	kmutex_t intrMux;

	/* fields for low-level interrupt handler, if needed */
	ddi_iblock_cookie_t softIblkCookie;
	kmutex_t lowMux;
	int softIntrRunning;
	ddi_softintr_t softIntrId;
	int stray;

	kmutex_t *muxp;
#define PMC_INTR_MUTEX_SET	0x01
#define PMC_LOW_MUTEX_SET	0x02
#define PMC_IHANDLER_SET	0x04
#define PMC_SOFT_IHANDLER_SET	0x08
#define PMC_REGS_MAPPED	0x10
#define PMC_MINOR_CREATED 0x20
	int resources;

	/* used to enforce exclusive open of channel devices */
	volatile char busy[PMC_NUM_CHANS];

#define PMC_OPEN	0x01
	volatile int flags;
	EnioRec enio;	/* for ATIIOC_ENABLE_IOPORTS */
	int iobase;
	int intrCnt;
} PmcSoftc;

#endif
