#include <stdio.h>
#include <sys/types.h>
#include <fcntl.h>
#include <errno.h>
#include <unistd.h>
#include <sys/mman.h>
#include <sys/file.h>

#include "s16dreg.h"
#include "s16dioctl.h"

#define MAINDEVNAME "/dev/RCIs16d0"
#define MMAPDEVNAME "/dev/RCIs16dmem0"

static S16dRegs *regsPtr;

main()
{
	/* Stuff for mapping the beast into memory */
	unsigned long regAddr = 0;
	int regSize = 20;
	int pageSize;
	off_t pageStart;
	caddr_t vpageStart;
	u_long pageOffset;
	
	u_short val;
	u_short val2;
	u_short set;

	int fd1;
	int fd2;

	if ((fd1 = open(MAINDEVNAME, O_RDWR)) < 0)
	 { perror ("open");
	   exit (1);
	 }
	if ((fd2 = open(MAINDEVNAME, O_RDWR)) >= 0)
	 { fprintf (stderr, "second open should not have worked\n");
	   exit (1);
	 }
	if (errno != EBUSY)
	 { perror ("second open");
	   exit (1);
	 }
	/* test the ioctls */

	if (ioctl (fd1, S16D_GET_COMMAND, &val) < 0)
	 { perror ("ioctl GET_COMMAND");
	   exit (1);
	 }
	printf ("command from ioctl: 0x%04x\n", val);
	val |= (S16D_FNCT1|S16D_FNCT2|S16D_FNCT3);
	set = val;
	if (ioctl (fd1, S16D_SET_COMMAND, &set) < 0)
	 { perror ("ioctl SET_COMMAND");
	   exit (1);
	 }
	ioctl (fd1, S16D_GET_COMMAND, &val2);
	if (val2 != val)
	 { fprintf (stderr, "Error, GET_COMMAND returns 0x%04x vs. 0x%04x\n",
		    val2, val);
	   exit (1);
	 }
	printf ("modified command from ioctl: 0x%04x\n", val);
	val &= ~(S16D_FNCT1|S16D_FNCT2|S16D_FNCT3);
	set = val;
	ioctl (fd1, S16D_SET_COMMAND, &set);

	if (ioctl (fd1, S16D_GET_STATUS, &val) < 0)
	 { perror ("ioctl GET_STATUS");
	   exit (1);
	 }
	printf ("status from ioctl: 0x%04x\n", val);
	if (ioctl (fd1, S16D_GET_CONFIG, &val) < 0)
	 { perror ("ioctl GET_CONFIG");
	   exit (1);
	 }
	printf ("config from ioctl: 0x%04x\n", val);
	val |= S16D_SWAP;
	set = val;
	if (ioctl (fd1, S16D_SET_CONFIG, &set) < 0)
	 { perror ("ioctl SET_CONFIG");
	   exit (1);
	 }
	ioctl (fd1, S16D_GET_CONFIG, &val2);
	if (val2 != val)
	 { fprintf (stderr, "Error, GET_CONFIG returns %x vs. %x\n",
		    val2, val);
	   exit (1);
	 }
	printf ("modified config from ioctl: 0x%04x\n", val);
	val &= ~(S16D_SWAP);
	set = val;
	ioctl (fd1, S16D_SET_CONFIG, &set);

	if ((fd2 = open(MMAPDEVNAME, O_RDWR)) < 0)
	 { perror ("third open");
	   exit (1);
	 }

	/* try to map in registers */

#ifdef __STDC__
    	pageSize = sysconf(_SC_PAGESIZE);
#else
    	pageSize = getpagesize();
#endif
	pageStart = (regAddr / pageSize) * pageSize;
	pageOffset = regAddr % pageSize;
	vpageStart = mmap((caddr_t)0, regSize, PROT_READ|PROT_WRITE,
			   MAP_SHARED, fd2, pageStart);
	if (vpageStart == (caddr_t)(-1))
	 { perror ("mmap");
	   exit (1);
	 }

	regsPtr = (S16dRegs *)(vpageStart + pageOffset);

	printf ("memory mapping:\n");
	printf ("stat: 0x%04x\n", regsPtr->csr);
	printf ("config: 0x%04x\n", regsPtr->config);
}
